from omuserver.server import Server
from omuserver.version import VERSION


def migrate(server: Server):
    version_path = server.config.directories.version
    if not version_path.exists():
        version_path.touch()
        version_path.write_text("0.0.0")
    previous_version = version_path.read_text().strip()
    if previous_version == "0.0.0":
        # Do something
        pass

    version_path.write_text(VERSION)



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from omu.extension.table.table_extension import (
    TABLE_ITEM_ADD_PACKET,
    TABLE_ITEM_CLEAR_PACKET,
    TABLE_ITEM_REMOVE_PACKET,
    TABLE_ITEM_UPDATE_PACKET,
    TableItemsPacket,
    TablePacket,
)
from omu.helper import batch_call
from omu.identifier import Identifier

from omuserver.extension.table.server_table import ServerTable
from omuserver.session import Session


class SessionTableListener:
    def __init__(self, id: Identifier, session: Session, table: ServerTable) -> None:
        self.id = id
        self.session = session
        self.table = table
        self.unlisten = batch_call(
            table.event.add.listen(self.on_add),
            table.event.update.listen(self.on_update),
            table.event.remove.listen(self.on_remove),
            table.event.clear.listen(self.on_clear),
        )

    def close(self) -> None:
        self.unlisten()

    async def on_add(self, items: Mapping[str, Any]) -> None:
        if self.session.closed:
            return
        await self.session.send(
            TABLE_ITEM_ADD_PACKET,
            TableItemsPacket(
                id=self.id,
                items=items,
            ),
        )

    async def on_update(self, items: Mapping[str, Any]) -> None:
        if self.session.closed:
            return
        await self.session.send(
            TABLE_ITEM_UPDATE_PACKET,
            TableItemsPacket(
                id=self.id,
                items=items,
            ),
        )

    async def on_remove(self, items: Mapping[str, Any]) -> None:
        if self.session.closed:
            return
        await self.session.send(
            TABLE_ITEM_REMOVE_PACKET,
            TableItemsPacket(
                id=self.id,
                items=items,
            ),
        )

    async def on_clear(self) -> None:
        if self.session.closed:
            return
        await self.session.send(TABLE_ITEM_CLEAR_PACKET, TablePacket(id=self.id))

    def __repr__(self) -> str:
        return f"<SessionTableHandler key={self.id} app={self.session.app}>"



from __future__ import annotations

import asyncio
import time

from omu.extension.dashboard.packets import PluginRequestPacket
from omu.extension.plugin import PackageInfo
from omu.extension.plugin.plugin_extension import (
    PLUGIN_ALLOWED_PACKAGE_TABLE,
    PLUGIN_REQUIRE_PACKET,
)
from packaging.specifiers import SpecifierSet

from omuserver.server import Server
from omuserver.session import Session
from omuserver.session.session import SessionType

from .plugin_loader import DependencyResolver, PluginLoader


class PluginExtension:
    def __init__(self, server: Server) -> None:
        self.server = server
        server.packet_dispatcher.register(
            PLUGIN_REQUIRE_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            PLUGIN_REQUIRE_PACKET,
            self.handle_require_packet,
        )
        server.network.event.start += self.on_network_start
        self.request_id = 0
        self.lock = asyncio.Lock()
        self.loader = PluginLoader(server)
        self.dependency_resolver = DependencyResolver()
        self.allowed_packages = server.tables.register(PLUGIN_ALLOWED_PACKAGE_TABLE)

    async def on_network_start(self) -> None:
        await self.loader.run_plugins()

    def _get_next_request_id(self) -> str:
        self.request_id += 1
        return f"{self.request_id}-{time.time_ns()}"

    async def request_plugins(
        self, session: Session, packages: dict[str, str | None]
    ) -> None:
        to_request: list[PackageInfo] = []
        for package in packages.keys():
            package_info = await self.dependency_resolver.get_installed_package_info(
                package
            )
            if package_info is None:
                package_info = await self.dependency_resolver.fetch_package_info(
                    package
                )
                to_request.append(package_info)
                continue
            await self.allowed_packages.add(package_info)
        if len(to_request) == 0:
            return
        request = PluginRequestPacket(
            request_id=self._get_next_request_id(),
            app=session.app,
            packages=to_request,
        )
        accepted = await self.server.dashboard.request_plugins(request)
        if not accepted:
            raise Exception("Request was not accepted")

    async def handle_require_packet(
        self, session: Session, packages: dict[str, str | None]
    ) -> None:
        if not packages:
            return

        async def task():
            if session.kind != SessionType.DASHBOARD:
                await self.request_plugins(session, packages)

            changed = False
            self.dependency_resolver.find_packages_distributions()
            for package, version in packages.items():
                specifier = None
                if version is not None:
                    specifier = SpecifierSet(version)
                if self.dependency_resolver.add_dependencies({package: specifier}):
                    changed = True

            if not changed:
                return

            async with self.lock:
                await self.dependency_resolver.resolve()
                await self.loader.load_updated_plugins()

        session.add_ready_task(task)



from __future__ import annotations

from collections.abc import Callable

from omu.errors import PermissionDenied
from omu.event_emitter import Unlisten
from omu.extension.signal import SignalPermissions
from omu.extension.signal.packets import SignalRegisterPacket
from omu.extension.signal.signal_extension import (
    SIGNAL_LISTEN_PACKET,
    SIGNAL_NOTIFY_PACKET,
    SIGNAL_REGISTER_PACKET,
    SignalPacket,
)
from omu.identifier import Identifier

from omuserver import Server
from omuserver.session import Session


class SignalExtension:
    def __init__(self, server: Server):
        self._server = server
        self.signals: dict[Identifier, ServerSignal] = {}
        server.packet_dispatcher.register(
            SIGNAL_REGISTER_PACKET,
            SIGNAL_LISTEN_PACKET,
            SIGNAL_NOTIFY_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            SIGNAL_REGISTER_PACKET, self.handle_register
        )
        server.packet_dispatcher.add_packet_handler(
            SIGNAL_LISTEN_PACKET, self.handle_listen
        )
        server.packet_dispatcher.add_packet_handler(
            SIGNAL_NOTIFY_PACKET, self.handle_notify
        )

    def get_signal(self, id: Identifier) -> ServerSignal:
        if id in self.signals:
            return self.signals[id]
        signal = ServerSignal(
            server=self._server,
            id=id,
            permissions=SignalPermissions(),
        )
        self.signals[id] = signal
        return signal

    def verify_permission(
        self,
        signal: ServerSignal,
        session: Session,
        get_scopes: Callable[[SignalPermissions], list[Identifier | None]],
    ) -> None:
        if signal.id.is_namepath_equal(session.app.id, path_length=1):
            return
        for permission in get_scopes(signal.permissions):
            if permission is None:
                continue
            if session.permission_handle.has(permission):
                return
        msg = f"App {session.app.id=} not allowed to access {signal.id=}"
        raise PermissionDenied(msg)

    async def handle_register(
        self, session: Session, data: SignalRegisterPacket
    ) -> None:
        if not data.id.is_subpath_of(session.app.id):
            raise PermissionDenied("App not allowed to register signal")
        signal = self.get_signal(data.id)
        signal.permissions = data.permissions

    async def handle_listen(self, session: Session, id: Identifier) -> None:
        signal = self.get_signal(id)
        self.verify_permission(
            signal,
            session,
            lambda permissions: [permissions.all, permissions.listen],
        )
        signal.attach_session(session)

    async def handle_notify(self, session: Session, data: SignalPacket) -> None:
        signal = self.get_signal(data.id)
        self.verify_permission(
            signal,
            session,
            lambda permissions: [permissions.all, permissions.notify],
        )
        await signal.notify(data.body)


class ServerSignal:
    def __init__(
        self,
        server: Server,
        id: Identifier,
        permissions: SignalPermissions,
    ) -> None:
        self.server = server
        self.id = id
        self.listeners: dict[Session, Unlisten] = {}
        self.permissions = permissions

    async def notify(self, body: bytes) -> None:
        packet = SignalPacket(id=self.id, body=body)
        for listener in self.listeners:
            await listener.send(SIGNAL_NOTIFY_PACKET, packet)

    def attach_session(self, session: Session) -> None:
        if session in self.listeners:
            raise Exception("Session already attached")
        unlisten = session.event.disconnected.listen(self.detach_session)
        self.listeners[session] = unlisten

    def detach_session(self, session: Session) -> None:
        if session not in self.listeners:
            raise Exception("Session not attached")
        unlisten = self.listeners.pop(session)
        unlisten()



import asyncio

from loguru import logger
from omu import Identifier
from omu.event_emitter import Unlisten
from omu.extension.registry.packets import RegistryPermissions
from omu.extension.registry.registry_extension import (
    REGISTRY_UPDATE_PACKET,
    RegistryPacket,
)
from omu.serializer import Serializable

from omuserver.server import Server
from omuserver.session import Session


class ServerRegistry:
    def __init__(
        self,
        server: Server,
        id: Identifier,
        permissions: RegistryPermissions | None = None,
    ) -> None:
        self.id = id
        self.permissions = permissions or RegistryPermissions()
        self._listeners: dict[Identifier, tuple[Session, Unlisten]] = {}
        self._path = server.directories.get(
            "registry"
        ) / id.get_sanitized_path().with_suffix(".json")
        self._changed = False
        self.value: bytes | None = None
        self.save_task: asyncio.Task | None = None

    async def load(self):
        if self._changed:
            raise Exception("Registry already loaded")
        if self._path.exists():
            self.value = self._path.read_bytes()

    async def store(self, value: bytes | None) -> None:
        self.value = value
        self._changed = True
        if self.save_task is None:
            self.save_task = asyncio.create_task(self._save())

    async def _save(self) -> None:
        while self._changed:
            if self.value is None:
                self._path.unlink(missing_ok=True)
            else:
                self._path.parent.mkdir(parents=True, exist_ok=True)
                self._path.write_bytes(self.value)
            self._changed = False
            await asyncio.sleep(1)
        self.save_task = None

    async def notify(self, session: Session) -> None:
        async with asyncio.TaskGroup() as tg:
            for listener, _ in self._listeners.values():
                if listener == session:
                    continue
                if listener.closed:
                    raise Exception(f"Session {listener.app=} closed")
                send = listener.send(
                    REGISTRY_UPDATE_PACKET,
                    RegistryPacket(id=self.id, value=self.value),
                )
                tg.create_task(send)

    async def attach_session(self, session: Session) -> None:
        if session.app.id in self._listeners:
            logger.warning(f"Session {session.app.id} is already to registry {self.id}")
            return
        unlisten = session.event.disconnected.listen(self.detach_session)
        self._listeners[session.app.id] = session, unlisten
        await session.send(
            REGISTRY_UPDATE_PACKET,
            RegistryPacket(id=self.id, value=self.value),
        )

    async def detach_session(self, session: Session) -> None:
        if session.app.id not in self._listeners:
            raise Exception("Session not attached")
        _, unlisten = self._listeners.pop(session.app.id)
        unlisten()


class Registry[T]:
    def __init__(
        self,
        registry: ServerRegistry,
        default_value: T,
        serializer: Serializable[T, bytes],
    ) -> None:
        self._registry = registry
        self._default_value = default_value
        self._serializer = serializer

    async def get(self) -> T:
        if self._registry.value is None:
            return self._default_value
        return self._serializer.deserialize(self._registry.value)

    async def set(self, value: T) -> None:
        await self._registry.store(self._serializer.serialize(value))



from omu.extension.logger.logger_extension import LOGGER_LOG_PERMISSION_ID
from omu.extension.permission.permission import PermissionType

LOGGER_LOG_PERMISSION = PermissionType(
    id=LOGGER_LOG_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "アプリの動作記録を保存",
            "en": "Save app operation records",
        },
        "note": {
            "ja": "不具合などの調査のために使われます",
            "en": "Used for investigation of problems",
        },
    },
)



from collections.abc import AsyncGenerator, Mapping

from omu.event_emitter import Unlisten
from omu.extension.table import Table, TableConfig, TableType
from omu.extension.table.table import TableEvents, TablePermissions
from omu.helper import AsyncCallback, Coro
from omu.identifier import Identifier
from omu.interface import Keyable
from omu.serializer import Serializable

from .server_table import ServerTable


class SerializeAdapter[T: Keyable](Mapping[str, T]):
    def __init__(self, cache: Mapping[str, bytes], serializer: Serializable[T, bytes]):
        self._cache = cache
        self._serializer = serializer

    def __getitem__(self, key: str) -> T:
        return self._serializer.deserialize(self._cache[key])


class SerializedTable[T: Keyable](Table[T]):
    def __init__(self, table: ServerTable, type: TableType[T]):
        self._table = table
        self._type = type
        self._event = TableEvents[T](self)
        self._proxies: list[Coro[[T], T | None]] = []
        self._chunk_size = 100
        self._permissions: TablePermissions | None = None
        self.permission_read: Identifier | None = None
        self.permission_write: Identifier | None = None
        self._listening = False
        table.event.cache_update += self.on_cache_update
        table.event.add += self.on_add
        table.event.update += self.on_update
        table.event.remove += self.on_remove
        table.event.clear += self.on_clear

    @property
    def cache(self) -> Mapping[str, T]:
        return SerializeAdapter(self._table.cache, self._type.serializer)

    def set_permissions(
        self,
        /,
        all: Identifier | None = None,
        read: Identifier | None = None,
        write: Identifier | None = None,
        remove: Identifier | None = None,
        proxy: Identifier | None = None,
    ) -> None:
        self._permissions = TablePermissions(
            all=all,
            read=read,
            write=write,
            remove=remove,
            proxy=proxy,
        )
        self._table.set_permissions(self._permissions)

    def set_config(self, config: TableConfig) -> None:
        self._table.set_config(config)

    def set_cache_size(self, size: int) -> None:
        self._table.set_cache_size(size)

    async def get(self, key: str) -> T | None:
        if key in self._table.cache:
            return self._type.serializer.deserialize(self._table.cache[key])
        item = await self._table.get(key)
        if item is None:
            return None
        return self._type.serializer.deserialize(item)

    async def get_many(self, *keys: str) -> dict[str, T]:
        items = await self._table.get_many(*keys)
        return {
            key: self._type.serializer.deserialize(item) for key, item in items.items()
        }

    async def add(self, *items: T) -> None:
        data = {item.key(): self._type.serializer.serialize(item) for item in items}
        await self._table.add(data)

    async def update(self, *items: T) -> None:
        data = {item.key(): self._type.serializer.serialize(item) for item in items}
        await self._table.update(data)

    async def remove(self, *items: T) -> None:
        await self._table.remove([item.key() for item in items])

    async def clear(self) -> None:
        await self._table.clear()

    async def has(self, key: str) -> bool:
        return await self._table.has(key)

    async def has_many(self, *keys: str) -> dict[str, bool]:
        return await self._table.has_many(list(keys))

    async def has_all(self, *keys: str) -> bool:
        return await self._table.has_all(list(keys))

    async def has_any(self, *keys: str) -> bool:
        return await self._table.has_any(list(keys))

    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> dict[str, T]:
        items = await self._table.fetch_items(limit, backward, cursor)
        return self._parse_items(items)

    async def fetch_range(self, start: str, end: str) -> dict[str, T]:
        items = await self._table.fetch_range(start, end)
        return self._parse_items(items)

    async def fetch_all(self) -> dict[str, T]:
        items = await self._table.fetch_all()
        return self._parse_items(items)

    async def iterate(
        self,
        backward: bool = False,
        cursor: str | None = None,
    ) -> AsyncGenerator[T, None]:
        while True:
            items = await self.fetch_items(
                limit=self._chunk_size,
                backward=backward,
                cursor=cursor,
            )
            if len(items) == 0:
                break
            for item in items.values():
                yield item
            cursor = next(iter(items.keys()))

    async def size(self) -> int:
        return await self._table.size()

    @property
    def event(self) -> TableEvents[T]:
        return self._event

    def listen(
        self, listener: AsyncCallback[Mapping[str, T]] | None = None
    ) -> Unlisten:
        self._listening = True
        if listener:
            return self._event.cache_update.listen(listener)
        return lambda: None

    async def on_cache_update(self, cache: Mapping[str, bytes]) -> None:
        await self._event.cache_update(self._parse_items(cache))

    async def on_add(self, items: Mapping[str, bytes]) -> None:
        _items = self._parse_items(items)
        await self._event.add(_items)

    async def on_update(self, items: Mapping[str, bytes]) -> None:
        _items = self._parse_items(items)
        await self._event.update(_items)

    async def on_remove(self, items: Mapping[str, bytes]) -> None:
        _items = self._parse_items(items)
        await self._event.remove(_items)

    async def on_clear(self) -> None:
        await self._event.clear()

    def proxy(self, callback: Coro[[T], T | None]) -> Unlisten:
        raise NotImplementedError

    def _parse_items(self, items: Mapping[str, bytes]) -> dict[str, T]:
        parsed: dict[str, T] = {}
        for key, item in items.items():
            item = self._type.serializer.deserialize(item)
            if not item:
                raise Exception(f"Failed to deserialize item {key}")
            parsed[key] = item
        return parsed



from __future__ import annotations

from aiohttp import web
from loguru import logger
from omu.bytebuffer import ByteReader, ByteWriter
from omu.network.packet import Packet, PacketData
from omu.network.packet_mapper import PacketMapper

from .session import SessionConnection


class WebsocketsConnection(SessionConnection):
    def __init__(self, socket: web.WebSocketResponse) -> None:
        self.socket = socket

    @property
    def closed(self) -> bool:
        return self.socket.closed

    async def receive(self, packet_mapper: PacketMapper) -> Packet | None:
        msg = await self.socket.receive()
        if msg.type in {
            web.WSMsgType.CLOSE,
            web.WSMsgType.CLOSING,
            web.WSMsgType.CLOSED,
        }:
            return None
        if msg.type != web.WSMsgType.BINARY:
            raise RuntimeError(f"Unknown message type {msg.type}")

        if msg.data is None:
            raise RuntimeError("Received empty message")
        if msg.type == web.WSMsgType.TEXT:
            raise RuntimeError("Received text message")
        elif msg.type == web.WSMsgType.BINARY:
            with ByteReader(msg.data) as reader:
                event_type = reader.read_string()
                event_data = reader.read_byte_array()
            packet_data = PacketData(event_type, event_data)
            return packet_mapper.deserialize(packet_data)
        else:
            raise RuntimeError(f"Unknown message type {msg.type}")

    async def close(self) -> None:
        try:
            await self.socket.close()
        except Exception as e:
            logger.warning(f"Error closing socket: {e}")
            logger.error(e)

    async def send(self, packet: Packet, packet_mapper: PacketMapper) -> None:
        if self.closed:
            raise ValueError("Socket is closed")
        packet_data = packet_mapper.serialize(packet)
        writer = ByteWriter()
        writer.write_string(packet_data.type)
        writer.write_byte_array(packet_data.data)
        await self.socket.send_bytes(writer.finish())



import abc
import datetime
import json
import random
import sqlite3
import string
from collections.abc import Iterable

from loguru import logger
from omu import App
from omu.extension.permission.permission import PermissionType
from omu.identifier import Identifier
from result import Ok, Result

from omuserver.server import Server
from omuserver.session import SessionType

type Token = str


class PermissionHandle(abc.ABC):
    @abc.abstractmethod
    def set_permissions(self, *permission_ids: Identifier) -> None: ...

    @abc.abstractmethod
    def has(self, permission_id: Identifier) -> bool: ...

    @abc.abstractmethod
    def has_any(self, permission_ids: Iterable[Identifier]) -> bool: ...

    @abc.abstractmethod
    def has_all(self, permission_ids: Iterable[Identifier]) -> bool: ...


class PermissionManager(abc.ABC):
    @abc.abstractmethod
    def register(self, *permission_types: PermissionType) -> None: ...

    @abc.abstractmethod
    def get_permission(self, permission_id: Identifier) -> PermissionType | None: ...

    @abc.abstractmethod
    def has_permission(self, token: Token, permission_id: Identifier) -> bool: ...

    @abc.abstractmethod
    def set_permissions(self, token: Token, *permission_ids: Identifier) -> None: ...

    @abc.abstractmethod
    async def generate_app_token(self, app: App) -> Token: ...

    @abc.abstractmethod
    async def validate_app_token(self, app: App, token: Token) -> bool: ...

    @abc.abstractmethod
    async def verify_app_token(
        self, app: App, token: Token | None
    ) -> Result[tuple[SessionType, PermissionHandle, Token], str]: ...

    @abc.abstractmethod
    def generate_plugin_token(self) -> Token: ...

    @abc.abstractmethod
    def is_plugin_token(self, token: Token) -> bool: ...

    @abc.abstractmethod
    def is_dashboard_token(self, token: Token) -> bool: ...


class TokenGenerator:
    def __init__(self):
        self._chars = string.ascii_letters + string.digits

    def generate(self, length: int) -> str:
        return "".join(random.choices(self._chars, k=length))


class ServerPermissionManager(PermissionManager):
    def __init__(self, server: Server):
        self._server = server
        self._plugin_tokens: set[str] = set()
        self._token_generator = TokenGenerator()
        self.permissions: dict[Identifier, PermissionType] = {}
        self.token_permissions: dict[str, list[Identifier]] = {}
        self._token_db = sqlite3.connect(
            server.directories.get("security") / "tokens.sqlite"
        )
        self._token_db.execute(
            """
            CREATE TABLE IF NOT EXISTS tokens (
                token TEXT PRIMARY KEY,
                identifier TEXT,
                created_at INTEGER,
                last_used_at INTEGER
            )
            """
        )
        self.token_permissions: dict[str, list[Identifier]] = {}
        permission_dir = server.directories.get("permissions")
        permission_dir.mkdir(parents=True, exist_ok=True)
        self.permission_db = sqlite3.connect(permission_dir / "permissions.db")
        self.permission_db.execute(
            """
            CREATE TABLE IF NOT EXISTS permissions (
                id TEXT PRIMARY KEY,
                value BLOB
            )
            """
        )
        self.permission_db.commit()
        self.load_permissions()

    def load_permissions(self) -> None:
        cursor = self.permission_db.cursor()
        cursor.execute("SELECT id, value FROM permissions")
        for row in cursor:
            token = row[0]
            permissions = json.loads(row[1])
            self.token_permissions[token] = [
                Identifier.from_key(key) for key in permissions
            ]

    def store_permissions(self) -> None:
        cursor = self.permission_db.cursor()
        for token, permissions in self.token_permissions.items():
            permission_keys = [permission.key() for permission in permissions]
            permissions = json.dumps(permission_keys)
            cursor.execute(
                "INSERT OR REPLACE INTO permissions VALUES (?, ?)",
                (token, permissions),
            )
        self.permission_db.commit()

    def set_permissions(self, token: Token, *permission_ids: Identifier) -> None:
        self.token_permissions[token] = list(permission_ids)
        self.store_permissions()

    def register(self, *permission_types: PermissionType) -> None:
        for permission in permission_types:
            if permission.id in self.permissions:
                raise ValueError(f"Permission {permission.id} already registered")
            self.permissions[permission.id] = permission

    def get_permission(self, permission_id: Identifier) -> PermissionType | None:
        return self.permissions.get(permission_id)

    def has_permission(self, token: Token, permission_id: Identifier) -> bool:
        permissions = self.token_permissions.get(token)
        if permissions is None:
            return False
        return permission_id in permissions

    async def generate_app_token(self, app: App) -> Token:
        token = self._token_generator.generate(32)
        self._token_db.execute(
            """
            INSERT INTO tokens (token, identifier, created_at, last_used_at)
            VALUES (?, ?, ?, ?)
            """,
            (
                token,
                app.id.key(),
                datetime.datetime.now(),
                datetime.datetime.now(),
            ),
        )
        self._token_db.commit()
        return token

    async def validate_app_token(self, app: App, token: Token) -> bool:
        if self._server.config.dashboard_token == token:
            return True
        cursor = self._token_db.execute(
            """
            SELECT token
            FROM tokens
            WHERE token = ? AND identifier = ?
            """,
            (token, app.id.key()),
        )
        result = cursor.fetchone()
        if result is None:
            return False
        self._token_db.execute(
            """
            UPDATE tokens
            SET last_used_at = ?
            WHERE token = ?
            """,
            (datetime.datetime.now(), token),
        )
        return True

    async def verify_app_token(
        self, app: App, token: str | None
    ) -> Result[tuple[SessionType, PermissionHandle, Token], str]:
        if token is None:
            token = await self.generate_app_token(app)
        else:
            if self.is_dashboard_token(token):
                return Ok((SessionType.DASHBOARD, DashboardPermissionHandle(), token))
            if self.is_plugin_token(token):
                return Ok((SessionType.PLUGIN, PluginPermissionHandle(), token))
        verified = await self.validate_app_token(app, token)
        if not verified:
            logger.warning(f"Invalid token: {token}")
            logger.info(f"Generating new token for {app}")
            token = await self.generate_app_token(app)
        return Ok((SessionType.APP, SessionPermissionHandle(self, token), token))

    def is_dashboard_token(self, token: Token) -> bool:
        dashboard_token = self._server.config.dashboard_token
        if dashboard_token is None:
            return False
        return dashboard_token == token

    def generate_plugin_token(self) -> Token:
        token = self._token_generator.generate(32)
        self._plugin_tokens.add(token)
        return token

    def is_plugin_token(self, token: Token) -> bool:
        return token in self._plugin_tokens


class SessionPermissionHandle(PermissionHandle):
    def __init__(self, security: PermissionManager, token: Token):
        self._security = security
        self._token = token

    def set_permissions(self, *permission_ids: Identifier) -> None:
        self._security.set_permissions(self._token, *permission_ids)

    def has(self, permission_id: Identifier) -> bool:
        return self._security.has_permission(self._token, permission_id)

    def has_any(self, permission_ids: Iterable[Identifier]) -> bool:
        return any(self.has(permission_id) for permission_id in permission_ids)

    def has_all(self, permission_ids: Iterable[Identifier]) -> bool:
        return all(self.has(permission_id) for permission_id in permission_ids)


class PluginPermissionHandle(PermissionHandle):
    def set_permissions(self, *permission_ids: Identifier) -> None:
        pass

    def has(self, permission_id: Identifier) -> bool:
        return True

    def has_any(self, permission_ids: Iterable[Identifier]) -> bool:
        return True

    def has_all(self, permission_ids: Iterable[Identifier]) -> bool:
        return True


class DashboardPermissionHandle(PermissionHandle):
    def set_permissions(self, *permission_ids: Identifier) -> None:
        pass

    def has(self, permission_id: Identifier) -> bool:
        return True

    def has_any(self, permission_ids: Iterable[Identifier]) -> bool:
        return True

    def has_all(self, permission_ids: Iterable[Identifier]) -> bool:
        return True



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from loguru import logger
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet import Packet, PacketType
from omu.network.packet_mapper import PacketMapper

from omuserver.session import Session


class ServerPacketDispatcher:
    def __init__(self):
        self.packet_mapper = PacketMapper()
        self._packet_listeners: dict[Identifier, PacketHandler] = {}

    async def process_connection(self, session: Session) -> None:
        session.event.packet += self.process_packet

    async def process_packet(self, session: Session, packet: Packet) -> None:
        listeners = self._packet_listeners.get(packet.type.id)
        if not listeners:
            logger.warning(f"Received unknown packet type {packet.type}")
            return
        if listeners.handler is None:
            raise ValueError(f"No handler for packet type {packet.type}")
        await listeners.handler(session, packet.data)

    def register(self, *types: PacketType) -> None:
        self.packet_mapper.register(*types)
        for type in types:
            if self._packet_listeners.get(type.id):
                raise ValueError(f"Packet id {type.id} already registered")
            self._packet_listeners[type.id] = PacketHandler(type)

    def add_packet_handler[T](
        self,
        packet_type: PacketType[T],
        handler: Coro[[Session, T], None] | None = None,
    ) -> Callable[[Coro[[Session, T], None]], None]:
        if not self._packet_listeners.get(packet_type.id):
            raise ValueError(f"Packet type {packet_type.id} not registered")

        def decorator(func: Coro[[Session, T], None]) -> None:
            self._packet_listeners[packet_type.id].handler = func

        if handler:
            decorator(handler)
        return decorator


@dataclass(slots=True)
class PacketHandler[T]:
    packet_type: PacketType[T]
    handler: Coro[[Session, T], None] | None = None



from omu.extension.permission import PermissionType
from omu.extension.server import (
    SERVER_APPS_READ_PERMISSION_ID,
    SERVER_SESSIONS_READ_PERMISSION_ID,
    SERVER_SHUTDOWN_PERMISSION_ID,
)
from omu.extension.server.server_extension import TRUSTED_ORIGINS_GET_PERMISSION_ID

SERVER_SHUTDOWN_PERMISSION = PermissionType(
    id=SERVER_SHUTDOWN_PERMISSION_ID,
    metadata={
        "level": "high",
        "name": {
            "ja": "サーバーをシャットダウン",
            "en": "Shutdown Server",
        },
        "note": {
            "ja": "アプリが内部のAPIサーバーをシャットダウンするために使われます",
            "en": "Used by apps to shut down the internal API server",
        },
    },
)
SERVER_APPS_READ_PERMISSION = PermissionType(
    id=SERVER_APPS_READ_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "アプリ一覧を取得",
            "en": "Get Running Apps",
        },
        "note": {
            "ja": "すべてのアプリ一覧を取得するために使われます",
            "en": "Used to get a list of apps connected to the server",
        },
    },
)
SERVER_SESSIONS_READ_PERMISSION = PermissionType(
    id=SERVER_SESSIONS_READ_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "接続中のアプリを取得",
            "en": "Get Running Apps",
        },
        "note": {
            "ja": "接続されているアプリ一覧を取得するために使われます",
            "en": "Used to get a list of apps connected to the server",
        },
    },
)
SERVER_TRUSTED_ORIGINS_GET_PERMISSION = PermissionType(
    id=TRUSTED_ORIGINS_GET_PERMISSION_ID,
    metadata={
        "level": "medium",
        "name": {
            "ja": "信頼されたオリジンを取得",
            "en": "Get Trusted Origins",
        },
        "note": {
            "ja": "認証を通過するオリジンを取得するために使われます",
            "en": "Used to get origins that pass authentication",
        },
    },
)



from __future__ import annotations

import sqlite3
from collections.abc import Mapping
from pathlib import Path

from .tableadapter import TableAdapter


class SqliteTableAdapter(TableAdapter):
    def __init__(self, path: Path) -> None:
        self._path = path
        self._conn = sqlite3.connect(path.with_suffix(".db"))
        self._conn.execute(
            # index, key, value
            "CREATE TABLE IF NOT EXISTS data ("
            "id INTEGER PRIMARY KEY AUTOINCREMENT,"
            "key TEXT UNIQUE,"
            "value BLOB"
            ")"
        )
        self._conn.commit()

    @classmethod
    def create(cls, path: Path) -> TableAdapter:
        return SqliteTableAdapter(path)

    async def store(self) -> None:
        pass

    async def load(self) -> None:
        pass

    async def get(self, key: str) -> bytes | None:
        cursor = self._conn.execute("SELECT value FROM data WHERE key = ?", (key,))
        row = cursor.fetchone()
        if row is None:
            return None
        return row[0]

    async def get_many(self, keys: list[str] | tuple[str, ...]) -> dict[str, bytes]:
        cursor = self._conn.execute(
            f"SELECT key, value FROM data WHERE key IN ({','.join('?' for _ in keys)})",
            keys,
        )
        rows = cursor.fetchall()
        return {row[0]: row[1] for row in rows}

    async def set(self, key: str, value: bytes) -> None:
        self._conn.execute(
            "INSERT OR REPLACE INTO data (key, value) VALUES (?, ?)",
            (key, value),
        )
        self._conn.commit()

    async def set_all(self, items: Mapping[str, bytes]) -> None:
        query = list(items.items())
        self._conn.executemany(
            "INSERT OR REPLACE INTO data (key, value) VALUES (?, ?)",
            query,
        )
        self._conn.commit()

    async def remove(self, key: str) -> None:
        self._conn.execute("DELETE FROM data WHERE key = ?", (key,))
        self._conn.commit()

    async def remove_all(self, keys: list[str] | tuple[str, ...]) -> None:
        self._conn.execute(
            f"DELETE FROM data WHERE key IN ({','.join('?' for _ in keys)})",
            keys,
        )
        self._conn.commit()

    async def has(self, key: str) -> bool:
        cursor = self._conn.execute("SELECT 1 FROM data WHERE key = ?", (key,))
        return cursor.fetchone() is not None

    async def has_many(self, keys: list[str] | tuple[str, ...]) -> dict[str, bool]:
        cursor = self._conn.execute(
            f"SELECT key FROM data WHERE key IN ({','.join('?' for _ in keys)})",
            keys,
        )
        rows = cursor.fetchall()
        return {row[0]: True for row in rows}

    async def has_any(self, keys: list[str] | tuple[str, ...]) -> bool:
        cursor = self._conn.execute(
            f"SELECT 1 FROM data WHERE key IN ({','.join('?' for _ in keys)})",
            keys,
        )
        return cursor.fetchone() is not None

    async def has_all(self, keys: list[str] | tuple[str, ...]) -> bool:
        cursor = self._conn.execute(
            f"SELECT key FROM data WHERE key IN ({','.join('?' for _ in keys)})",
            keys,
        )
        rows = cursor.fetchall()
        return len(rows) == len(keys)

    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> dict[str, bytes]:
        cursor_id: int | None = None
        if cursor is None:
            if backward:
                _cursor = self._conn.execute(
                    "SELECT id FROM data ORDER BY id DESC LIMIT 1"
                )
            else:
                _cursor = self._conn.execute("SELECT id FROM data ORDER BY id LIMIT 1")
            row = _cursor.fetchone()
            if row is None:
                return {}
            cursor_id = row[0]
        else:
            _cursor = self._conn.execute("SELECT id FROM data WHERE key = ?", (cursor,))
            row = _cursor.fetchone()
            if row is None:
                raise ValueError(f"Cursor {cursor} not found")
            cursor_id = row[0]

        if backward:
            _cursor = self._conn.execute(
                "SELECT key, value FROM data WHERE id <= ? ORDER BY id DESC LIMIT ?",
                (cursor_id, limit),
            )
        else:
            _cursor = self._conn.execute(
                "SELECT key, value FROM data WHERE id >= ? ORDER BY id LIMIT ?",
                (cursor_id, limit),
            )
        return {row[0]: (row[1]) for row in _cursor.fetchall()}

    async def fetch_range(self, start: str, end: str) -> dict[str, bytes]:
        start_id: int
        end_id: int
        _cursor = self._conn.execute("SELECT id FROM data WHERE key = ?", (start,))
        row = _cursor.fetchone()
        if row is None:
            raise ValueError(f"start key {start} not found")
        start_id = row[0]

        _cursor = self._conn.execute("SELECT id FROM data WHERE key = ?", (end,))
        row = _cursor.fetchone()
        if row is None:
            raise ValueError(f"end key {end} not found")
        end_id = row[0]

        _cursor = self._conn.execute(
            "SELECT key, value FROM data WHERE id >= ? AND id <= ?",
            (start_id, end_id),
        )
        return {row[0]: (row[1]) for row in _cursor.fetchall()}

    async def fetch_all(self) -> dict[str, bytes]:
        _cursor = self._conn.execute("SELECT key, value FROM data")
        return {row[0]: (row[1]) for row in _cursor.fetchall()}

    async def first(self) -> str | None:
        _cursor = self._conn.execute("SELECT key FROM data ORDER BY id LIMIT 1")
        row = _cursor.fetchone()
        if row is None:
            return None
        return row[0]

    async def last(self) -> str | None:
        _cursor = self._conn.execute("SELECT key FROM data ORDER BY id DESC LIMIT 1")
        row = _cursor.fetchone()
        if row is None:
            return None
        return row[0]

    async def clear(self) -> None:
        self._conn.execute("DELETE FROM data")
        self._conn.commit()

    async def size(self) -> int:
        _cursor = self._conn.execute("SELECT COUNT(*) FROM data")
        row = _cursor.fetchone()
        if row is None:
            return 0
        return row[0]



from __future__ import annotations

import abc
import asyncio
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from loguru import logger
from omu import App
from omu.errors import DisconnectReason
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.network.packet import PACKET_TYPES, Packet, PacketType
from omu.network.packet.packet_types import (
    ConnectPacket,
    DisconnectPacket,
    DisconnectType,
)
from omu.network.packet_mapper import PacketMapper
from result import Err, Ok

from omuserver.server import Server

if TYPE_CHECKING:
    from omuserver.security import PermissionHandle


class SessionConnection(abc.ABC):
    @abc.abstractmethod
    async def send(self, packet: Packet, packet_mapper: PacketMapper) -> None: ...

    @abc.abstractmethod
    async def receive(self, packet_mapper: PacketMapper) -> Packet | None: ...

    @abc.abstractmethod
    async def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool: ...


class SessionEvents:
    def __init__(self) -> None:
        self.packet = EventEmitter[Session, Packet]()
        self.disconnected = EventEmitter[Session](catch_errors=True)
        self.ready = EventEmitter[Session]()


@dataclass(frozen=True, slots=True)
class SessionTask:
    session: Session
    coro: Coro[[], None]
    name: str


class SessionType(Enum):
    APP = "app"
    PLUGIN = "plugin"
    DASHBOARD = "dashboard"


class Session:
    def __init__(
        self,
        packet_mapper: PacketMapper,
        app: App,
        permission_handle: PermissionHandle,
        kind: SessionType,
        connection: SessionConnection,
    ) -> None:
        self.packet_mapper = packet_mapper
        self.app = app
        self.permission_handle = permission_handle
        self.kind = kind
        self.connection = connection
        self.event = SessionEvents()
        self.ready_tasks: list[SessionTask] = []
        self.ready = False

    @classmethod
    async def from_connection(
        cls,
        server: Server,
        packet_mapper: PacketMapper,
        connection: SessionConnection,
    ) -> Session:
        packet = await connection.receive(packet_mapper)
        if packet is None:
            await connection.close()
            raise RuntimeError("Connection closed")
        if packet.type != PACKET_TYPES.CONNECT:
            await connection.send(
                Packet(
                    PACKET_TYPES.DISCONNECT,
                    DisconnectPacket(
                        DisconnectType.INVALID_PACKET_TYPE, "Expected connect"
                    ),
                ),
                packet_mapper,
            )
            await connection.close()
            raise RuntimeError(
                f"Expected {PACKET_TYPES.CONNECT.id} but got {packet.type}"
            )
        if not isinstance(packet.data, ConnectPacket):
            await connection.send(
                Packet(
                    PACKET_TYPES.DISCONNECT,
                    DisconnectPacket(
                        DisconnectType.INVALID_PACKET_TYPE, "Expected connect"
                    ),
                ),
                packet_mapper,
            )
            await connection.close()
            raise RuntimeError(f"Invalid packet data: {packet.data}")
        event = packet.data
        app = event.app
        token = event.token

        match await server.permission_manager.verify_app_token(app, token):
            case Ok((kind, permission_handle, new_token)):
                session = Session(
                    packet_mapper=packet_mapper,
                    app=app,
                    permission_handle=permission_handle,
                    kind=kind,
                    connection=connection,
                )
                if session.kind != SessionType.PLUGIN:
                    await session.send(PACKET_TYPES.TOKEN, new_token)
                return session
            case Err(error):
                await connection.send(
                    Packet(
                        PACKET_TYPES.DISCONNECT,
                        DisconnectPacket(DisconnectType.INVALID_TOKEN, error),
                    ),
                    packet_mapper,
                )
                await connection.close()
                raise RuntimeError(f"Invalid token: {error}")

    @property
    def closed(self) -> bool:
        return self.connection.closed

    async def disconnect(
        self, disconnect_type: DisconnectType, message: str | None = None
    ) -> None:
        if not self.connection.closed:
            await self.send(
                PACKET_TYPES.DISCONNECT, DisconnectPacket(disconnect_type, message)
            )
        await self.connection.close()
        await self.event.disconnected.emit(self)

    async def listen(self) -> None:
        while not self.connection.closed:
            packet = await self.connection.receive(self.packet_mapper)
            if packet is None:
                await self.disconnect(DisconnectType.CLOSE)
                return
            asyncio.create_task(self.dispatch_packet(packet))

    async def dispatch_packet(self, packet: Packet) -> None:
        try:
            await self.event.packet.emit(self, packet)
        except DisconnectReason as reason:
            logger.opt(exception=reason).error("Disconnecting session")
            await self.disconnect(reason.type, reason.message)

    async def send[T](self, packet_type: PacketType[T], data: T) -> None:
        await self.connection.send(Packet(packet_type, data), self.packet_mapper)

    def add_ready_task(self, coro: Coro[[], None]):
        if self.ready:
            raise RuntimeError("Session is already ready")
        task = SessionTask(
            session=self,
            coro=coro,
            name=coro.__name__,
        )
        self.ready_tasks.append(task)

    async def process_ready_tasks(self) -> None:
        if self.ready:
            raise RuntimeError("Session is already ready")
        for task in self.ready_tasks:
            await task.coro()
        self.ready_tasks.clear()
        self.ready = True
        await self.event.ready.emit(self)



from __future__ import annotations

from collections.abc import Callable

from omu.errors import PermissionDenied
from omu.extension.permission import PermissionType
from omu.extension.registry import RegistryType
from omu.extension.registry.packets import RegistryPermissions, RegistryRegisterPacket
from omu.extension.registry.registry_extension import (
    REGISTRY_GET_ENDPOINT,
    REGISTRY_LISTEN_PACKET,
    REGISTRY_PERMISSION_ID,
    REGISTRY_REGISTER_PACKET,
    REGISTRY_UPDATE_PACKET,
    RegistryPacket,
)
from omu.identifier import Identifier

from omuserver.server import Server
from omuserver.session import Session

from .registry import Registry, ServerRegistry

REGISTRY_PERMISSION = PermissionType(
    REGISTRY_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "レジストリ",
            "en": "Registry Permission",
        },
        "note": {
            "ja": "アプリがデータを保持するために使われます",
            "en": "Used by apps to store data",
        },
    },
)


class RegistryExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        self.registries: dict[Identifier, ServerRegistry] = {}
        self._startup_registries: list[ServerRegistry] = []
        server.permission_manager.register(REGISTRY_PERMISSION)
        server.packet_dispatcher.register(
            REGISTRY_REGISTER_PACKET,
            REGISTRY_LISTEN_PACKET,
            REGISTRY_UPDATE_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            REGISTRY_REGISTER_PACKET, self.handle_register
        )
        server.packet_dispatcher.add_packet_handler(
            REGISTRY_LISTEN_PACKET, self.handle_listen
        )
        server.packet_dispatcher.add_packet_handler(
            REGISTRY_UPDATE_PACKET, self.handle_update
        )
        server.endpoints.bind_endpoint(REGISTRY_GET_ENDPOINT, self.handle_get)
        server.event.start += self._on_start

    async def _on_start(self) -> None:
        for registry in self._startup_registries:
            await registry.load()
        self._startup_registries.clear()

    async def handle_register(
        self, session: Session, packet: RegistryRegisterPacket
    ) -> None:
        registry = await self.get(packet.id)
        if not registry.id.is_subpath_of(session.app.id):
            msg = f"App {session.app.id=} not allowed to register {packet.id=}"
            raise PermissionDenied(msg)
        registry.permissions = packet.permissions

    async def handle_listen(self, session: Session, id: Identifier) -> None:
        registry = await self.get(id)
        self.verify_permission(
            registry,
            session,
            lambda permissions: [permissions.all, permissions.read],
        )
        await registry.attach_session(session)

    async def handle_update(self, session: Session, packet: RegistryPacket) -> None:
        registry = await self.get(packet.id)
        self.verify_permission(
            registry,
            session,
            lambda permissions: [permissions.all, permissions.write],
        )
        await registry.store(packet.value)
        await registry.notify(session)

    async def handle_get(self, session: Session, id: Identifier) -> RegistryPacket:
        registry = await self.get(id)
        self.verify_permission(
            registry,
            session,
            lambda permissions: [permissions.all, permissions.read],
        )
        return RegistryPacket(id, registry.value)

    async def get(self, id: Identifier) -> ServerRegistry:
        registry = self.registries.get(id)
        if registry is None:
            registry = ServerRegistry(
                server=self._server,
                id=id,
            )
            self.registries[id] = registry
            await registry.load()
        return registry

    def verify_permission(
        self,
        registry: ServerRegistry,
        session: Session,
        get_scopes: Callable[[RegistryPermissions], list[Identifier | None]],
    ) -> None:
        if registry.id.is_namepath_equal(session.app.id, path_length=1):
            return
        require_permissions = get_scopes(registry.permissions)
        # if not any(
        #     self._server.permissions.has_permission(session, permission)
        #     for permission in filter(None, require_permissions)
        # ):
        #     msg = f"App {session.app.id=} not allowed to access {registry.id=}"
        #     raise PermissionDenied(msg)
        if not session.permission_handle.has_any(filter(None, require_permissions)):
            msg = f"App {session.app.id=} not allowed to access {registry.id=}"
            raise PermissionDenied(msg)

    def register[T](
        self,
        registry_type: RegistryType[T],
    ) -> Registry[T]:
        registry = self.registries.get(registry_type.id)
        if registry is None:
            registry = ServerRegistry(
                server=self._server,
                id=registry_type.id,
                permissions=registry_type.permissions,
            )
            self.registries[registry_type.id] = registry
            self._startup_registries.append(registry)
        return Registry(
            registry,
            registry_type.default_value,
            registry_type.serializer,
        )

    async def store(self, id: Identifier, value: bytes) -> None:
        registry = await self.get(id)
        await registry.store(value)



def test_directory_traversal_attack():
    from pathlib import Path

    from omuserver.helper import safe_path

    root = Path("/home/omu")
    path = Path("/home/omu/../../etc/passwd")
    try:
        safe_path(root, path)
    except ValueError:
        pass
    else:
        raise AssertionError("Expected ValueError")

    root = Path("/home/omu")
    path = Path("/home/omu/etc/passwd")
    safe_path(root, path)



from omu.extension.i18n import I18N_SET_LOCALES_PERMISSION_ID
from omu.extension.i18n.i18n_extension import I18N_GET_LOCALES_PERMISSION_ID
from omu.extension.permission import PermissionType

I18N_SET_LOCALES_PERMISSION = PermissionType(
    id=I18N_SET_LOCALES_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "地域設定を変更",
            "en": "Change locale settings",
        },
        "note": {
            "ja": "言語や通貨など地域設定を変更するために使われます",
            "en": "Used to change locale settings such as language and currency",
        },
    },
)
I18N_GET_LOCALES_PERMISSION = PermissionType(
    id=I18N_GET_LOCALES_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "地域設定を取得",
            "en": "Get locale settings",
        },
        "note": {
            "ja": "言語や通貨など地域設定を取得するために使われます",
            "en": "Used to get locale settings such as language and currency",
        },
    },
)



from loguru import logger
from omu.extension.logger.logger_extension import (
    LOGGER_LISTEN_PACKET,
    LOGGER_LOG_PACKET,
)
from omu.extension.logger.packets import LogMessage, LogPacket
from omu.identifier import Identifier

from omuserver.server import Server
from omuserver.session import Session

from .permissions import (
    LOGGER_LOG_PERMISSION,
)


class LoggerExtension:
    def __init__(self, server: Server):
        server.permission_manager.register(
            LOGGER_LOG_PERMISSION,
        )
        server.packet_dispatcher.register(
            LOGGER_LOG_PACKET,
            LOGGER_LISTEN_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            LOGGER_LOG_PACKET,
            self.handle_log,
        )
        server.packet_dispatcher.add_packet_handler(
            LOGGER_LISTEN_PACKET, self.handle_listen
        )
        self.listeners: dict[Identifier, set[Session]] = {}

    async def broadcast(self, id: Identifier, message: LogMessage) -> None:
        packet = LogPacket(id=id, message=message)
        for session in self.listeners.get(id, []):
            await session.send(LOGGER_LOG_PACKET, packet)

    async def handle_log(self, session: Session, packet: LogPacket) -> None:
        logger.info(f"{packet.id}: {packet.message}")

    async def handle_listen(self, session: Session, id: Identifier) -> None:
        if id not in self.listeners:
            self.listeners[id] = set()
        self.listeners[id].add(session)



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path

from omu.errors import PermissionDenied
from omu.extension.permission import PermissionType
from omu.extension.table import Table, TablePermissions, TableType
from omu.extension.table.packets import (
    SetConfigPacket,
    SetPermissionPacket,
    TableFetchPacket,
    TableFetchRangePacket,
    TableItemsPacket,
    TableKeysPacket,
    TablePacket,
    TableProxyPacket,
)
from omu.extension.table.table_extension import (
    TABLE_FETCH_ALL_ENDPOINT,
    TABLE_FETCH_ENDPOINT,
    TABLE_FETCH_RANGE_ENDPOINT,
    TABLE_ITEM_ADD_PACKET,
    TABLE_ITEM_CLEAR_PACKET,
    TABLE_ITEM_GET_ENDPOINT,
    TABLE_ITEM_HAS_ALL_ENDPOINT,
    TABLE_ITEM_HAS_ANY_ENDPOINT,
    TABLE_ITEM_HAS_ENDPOINT,
    TABLE_ITEM_REMOVE_PACKET,
    TABLE_ITEM_UPDATE_PACKET,
    TABLE_LISTEN_PACKET,
    TABLE_PERMISSION_ID,
    TABLE_PROXY_LISTEN_PACKET,
    TABLE_PROXY_PACKET,
    TABLE_SET_CONFIG_PACKET,
    TABLE_SET_PERMISSION_PACKET,
    TABLE_SIZE_ENDPOINT,
)
from omu.identifier import Identifier
from omu.interface import Keyable

from omuserver.server import Server
from omuserver.session import Session

from .adapters.sqlitetable import SqliteTableAdapter
from .adapters.tableadapter import TableAdapter
from .cached_table import CachedTable
from .serialized_table import SerializedTable
from .server_table import ServerTable

TABLE_PERMISSION = PermissionType(
    TABLE_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "テーブル",
            "en": "Table Permission",
        },
        "note": {
            "ja": "テーブル形式のデータを扱うために使われます",
            "en": "Used to handle table formatted data",
        },
    },
)


class TableExtension:
    def __init__(self, server: Server) -> None:
        self.server = server
        self._tables: dict[Identifier, ServerTable] = {}
        self._adapters: list[TableAdapter] = []
        server.permission_manager.register(TABLE_PERMISSION)
        server.packet_dispatcher.register(
            TABLE_SET_PERMISSION_PACKET,
            TABLE_SET_CONFIG_PACKET,
            TABLE_LISTEN_PACKET,
            TABLE_PROXY_LISTEN_PACKET,
            TABLE_PROXY_PACKET,
            TABLE_ITEM_ADD_PACKET,
            TABLE_ITEM_UPDATE_PACKET,
            TABLE_ITEM_REMOVE_PACKET,
            TABLE_ITEM_CLEAR_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_SET_PERMISSION_PACKET,
            self.handle_bind_permission,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_SET_CONFIG_PACKET,
            self.handle_table_config,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_LISTEN_PACKET,
            self.handler_listen,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_PROXY_LISTEN_PACKET,
            self.handle_proxy_listen,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_PROXY_PACKET,
            self.handle_proxy,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_ITEM_ADD_PACKET,
            self.handle_item_add,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_ITEM_UPDATE_PACKET,
            self.handle_item_update,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_ITEM_REMOVE_PACKET,
            self.handle_item_remove,
        )
        server.packet_dispatcher.add_packet_handler(
            TABLE_ITEM_CLEAR_PACKET,
            self.handle_item_clear,
        )
        server.endpoints.bind_endpoint(
            TABLE_ITEM_GET_ENDPOINT,
            self.handle_item_get,
        )
        server.endpoints.bind_endpoint(
            TABLE_ITEM_HAS_ENDPOINT,
            self.handle_item_has,
        )
        server.endpoints.bind_endpoint(
            TABLE_ITEM_HAS_ALL_ENDPOINT,
            self.handle_item_has_all,
        )
        server.endpoints.bind_endpoint(
            TABLE_ITEM_HAS_ANY_ENDPOINT,
            self.handle_item_has_any,
        )
        server.endpoints.bind_endpoint(
            TABLE_FETCH_ENDPOINT,
            self.handle_item_fetch,
        )
        server.endpoints.bind_endpoint(
            TABLE_FETCH_RANGE_ENDPOINT,
            self.handle_item_fetch_range,
        )
        server.endpoints.bind_endpoint(
            TABLE_FETCH_ALL_ENDPOINT,
            self.handle_item_fetch_all,
        )
        server.endpoints.bind_endpoint(
            TABLE_SIZE_ENDPOINT,
            self.handle_table_size,
        )
        server.event.start += self.on_server_start
        server.event.stop += self.on_server_stop

    async def handle_item_get(
        self, session: Session, packet: TableKeysPacket
    ) -> TableItemsPacket:
        table = await self.get_table(packet.id)
        items = await table.get_many(*packet.keys)
        return TableItemsPacket(
            id=packet.id,
            items=items,
        )

    async def handle_item_has(
        self, session: Session, packet: TableKeysPacket
    ) -> dict[str, bool]:
        table = await self.get_table(packet.id)
        return await table.has_many(packet.keys)

    async def handle_item_has_all(
        self, session: Session, packet: TableKeysPacket
    ) -> bool:
        table = await self.get_table(packet.id)
        return await table.has_all(packet.keys)

    async def handle_item_has_any(
        self, session: Session, packet: TableKeysPacket
    ) -> bool:
        table = await self.get_table(packet.id)
        return await table.has_any(packet.keys)

    async def handle_item_fetch(
        self, session: Session, packet: TableFetchPacket
    ) -> TableItemsPacket:
        table = await self.get_table(packet.id)
        items = await table.fetch_items(
            limit=packet.limit,
            backward=packet.backward,
            cursor=packet.cursor,
        )
        return TableItemsPacket(
            id=packet.id,
            items=items,
        )

    async def handle_item_fetch_range(
        self, session: Session, packet: TableFetchRangePacket
    ) -> TableItemsPacket:
        table = await self.get_table(packet.id)
        items = await table.fetch_range(packet.start, packet.end)
        return TableItemsPacket(
            id=packet.id,
            items=items,
        )

    async def handle_item_fetch_all(
        self, session: Session, packet: TablePacket
    ) -> TableItemsPacket:
        table = await self.get_table(packet.id)
        items = await table.fetch_all()
        return TableItemsPacket(
            id=packet.id,
            items=items,
        )

    async def handle_table_size(self, session: Session, packet: TablePacket) -> int:
        table = await self.get_table(packet.id)
        return await table.size()

    async def handle_bind_permission(
        self, session: Session, packet: SetPermissionPacket
    ) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all],
        )
        permissions = TablePermissions(
            all=packet.all,
            read=packet.read,
            write=packet.write,
            remove=packet.remove,
            proxy=packet.proxy,
        )
        table.set_permissions(permissions)

    async def handle_table_config(
        self, session: Session, packet: SetConfigPacket
    ) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all],
        )
        table.set_config(packet.config)

    async def handler_listen(self, session: Session, id: Identifier) -> None:
        table = await self.get_table(id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.read],
        )
        table.attach_session(session)

    async def handle_proxy_listen(self, session: Session, id: Identifier) -> None:
        table = await self.get_table(id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.proxy],
        )
        table.attach_proxy_session(session)

    async def handle_proxy(self, session: Session, packet: TableProxyPacket) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.proxy],
        )
        await table.proxy(session, packet.key, packet.items)

    async def handle_item_add(self, session: Session, packet: TableItemsPacket) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.write],
        )
        await table.add(packet.items)

    async def handle_item_update(
        self, session: Session, packet: TableItemsPacket
    ) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.write],
        )
        await table.update(packet.items)

    async def handle_item_remove(
        self, session: Session, packet: TableItemsPacket
    ) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.remove],
        )
        await table.remove(list(packet.items.keys()))

    async def handle_item_clear(self, session: Session, packet: TablePacket) -> None:
        table = await self.get_table(packet.id)
        await self.verify_permission(
            session,
            table,
            lambda perms: [perms.all, perms.remove],
        )
        await table.clear()

    async def register_table[T: Keyable](self, table_type: TableType[T]) -> Table[T]:
        table = await self.get_table(table_type.id)
        return SerializedTable(table, table_type)

    async def get_table(self, id: Identifier) -> ServerTable:
        if id in self._tables:
            return self._tables[id]
        table = CachedTable(self.server, id)
        adapter = SqliteTableAdapter.create(self.get_table_path(id))
        await adapter.load()
        table.set_adapter(adapter)
        self._tables[id] = table
        return table

    def get_table_path(self, id: Identifier) -> Path:
        path = self.server.directories.get("tables") / id.get_sanitized_path()
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    async def on_server_start(self) -> None:
        for table in self._tables.values():
            await table.load()

    async def on_server_stop(self) -> None:
        for table in self._tables.values():
            await table.store()

    async def verify_permission(
        self,
        session: Session,
        table: ServerTable,
        get_scopes: Callable[[TablePermissions], list[Identifier | None]],
    ):
        if table.id.is_namepath_equal(session.app.id, path_length=1):
            return
        if table.permissions is None:
            raise PermissionDenied(f"Table {table.id} does not have a permission set")
        permissions: list[Identifier | None] = get_scopes(table.permissions)
        for permission in permissions:
            if permission is None:
                continue
            if session.permission_handle.has(permission):
                return

        raise PermissionDenied(
            f"Table {table.id} does not have permission {table.permissions}"
        )

    def register[T: Keyable](self, table_type: TableType[T]) -> Table[T]:
        table = CachedTable(self.server, table_type.id)
        table.set_permissions(table_type.permissions)
        adapter = SqliteTableAdapter.create(self.get_table_path(table_type.id))
        table.set_adapter(adapter)
        self._tables[table_type.id] = table
        return SerializedTable(table, table_type)



from __future__ import annotations

import asyncio

from omu.network.connection import Connection
from omu.network.packet import Packet, PacketData
from omu.serializer import Serializable


class PluginConnection(Connection):
    def __init__(self) -> None:
        self._connected = False
        self._to_client_queue = asyncio.Queue[Packet]()
        self._to_server_queue = asyncio.Queue[Packet]()

    async def connect(self) -> None:
        self._connected = True

    async def receive(self, packet_mapper: Serializable[Packet, PacketData]) -> Packet:
        return await self._to_client_queue.get()

    def add_receive(self, packet: Packet) -> None:
        self._to_client_queue.put_nowait(packet)

    async def send(
        self, packet: Packet, packet_mapper: Serializable[Packet, PacketData]
    ) -> None:
        self._to_server_queue.put_nowait(packet)

    async def dequeue_to_server_packet(self) -> Packet:
        return await self._to_server_queue.get()

    @property
    def closed(self) -> bool:
        return not self._connected

    async def close(self) -> None:
        self._connected = False



from __future__ import annotations

from loguru import logger
from omu.network import Packet
from omu.network.packet_mapper import PacketMapper

from omuserver.session import SessionConnection

from .plugin_connection import PluginConnection


class PluginSessionConnection(SessionConnection):
    def __init__(self, connection: PluginConnection) -> None:
        self.connection = connection

    @property
    def closed(self) -> bool:
        return self.connection.closed

    async def receive(self, packet_mapper: PacketMapper) -> Packet:
        return await self.connection.dequeue_to_server_packet()

    async def close(self) -> None:
        try:
            await self.connection.close()
        except Exception as e:
            logger.warning(f"Error closing socket: {e}")
            logger.error(e)

    async def send(self, packet: Packet, packet_mapper: PacketMapper) -> None:
        if self.closed:
            raise ValueError("Socket is closed")
        self.connection.add_receive(packet)

    def __repr__(self) -> str:
        return f"PluginSessionConnection({self.connection})"



from __future__ import annotations

import abc
from collections.abc import AsyncGenerator, Mapping

from omu.event_emitter import EventEmitter
from omu.extension.table import TableConfig
from omu.extension.table.table import TablePermissions
from omu.identifier import Identifier

from omuserver.session import Session

from .adapters.tableadapter import TableAdapter

type Json = str | int | float | bool | None | dict[str, Json] | list[Json]


class ServerTable(abc.ABC):
    @abc.abstractmethod
    async def load(self) -> None: ...

    @abc.abstractmethod
    async def store(self) -> None: ...

    @property
    @abc.abstractmethod
    def id(self) -> Identifier: ...

    @property
    @abc.abstractmethod
    def cache(self) -> Mapping[str, bytes]: ...

    @abc.abstractmethod
    def set_config(self, config: TableConfig) -> None: ...

    @property
    @abc.abstractmethod
    def permissions(self) -> TablePermissions | None: ...

    @abc.abstractmethod
    def set_permissions(self, permissions: TablePermissions) -> None: ...

    @abc.abstractmethod
    def set_cache_size(self, size: int) -> None: ...

    @property
    @abc.abstractmethod
    def adapter(self) -> TableAdapter | None: ...

    @abc.abstractmethod
    def set_adapter(self, adapter: TableAdapter) -> None: ...

    @abc.abstractmethod
    def attach_session(self, session: Session) -> None: ...

    @abc.abstractmethod
    def detach_session(self, session: Session) -> None: ...

    @abc.abstractmethod
    def attach_proxy_session(self, session: Session) -> None: ...

    @abc.abstractmethod
    async def proxy(
        self, session: Session, key: int, items: Mapping[str, bytes]
    ) -> int: ...

    @abc.abstractmethod
    async def get(self, key: str) -> bytes | None: ...

    @abc.abstractmethod
    async def get_many(self, *keys: str) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def add(self, items: Mapping[str, bytes]) -> None: ...

    @abc.abstractmethod
    async def update(self, items: Mapping[str, bytes]) -> None: ...

    @abc.abstractmethod
    async def remove(self, keys: list[str] | tuple[str, ...]) -> None: ...

    @abc.abstractmethod
    async def clear(self) -> None: ...

    @abc.abstractmethod
    async def has(self, key: str) -> bool: ...

    @abc.abstractmethod
    async def has_many(self, keys: list[str] | tuple[str, ...]) -> dict[str, bool]: ...

    @abc.abstractmethod
    async def has_all(self, keys: list[str] | tuple[str, ...]) -> bool: ...

    @abc.abstractmethod
    async def has_any(self, keys: list[str] | tuple[str, ...]) -> bool: ...

    @abc.abstractmethod
    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def fetch_range(self, start: str, end: str) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def fetch_all(self) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def iterate(self) -> AsyncGenerator[bytes, None]: ...

    @abc.abstractmethod
    async def size(self) -> int: ...

    @property
    @abc.abstractmethod
    def event(self) -> ServerTableEvents: ...


class ServerTableEvents:
    def __init__(self) -> None:
        self.add = EventEmitter[Mapping[str, bytes]]()
        self.update = EventEmitter[Mapping[str, bytes]]()
        self.remove = EventEmitter[Mapping[str, bytes]]()
        self.clear = EventEmitter[[]]()
        self.cache_update = EventEmitter[Mapping[str, bytes]]()



from __future__ import annotations

import asyncio
import importlib
import importlib.metadata
import io
import os
import sys
from dataclasses import dataclass
from multiprocessing import Process

from loguru import logger
from omu.address import Address
from omu.app import App
from omu.helper import asyncio_error_logger
from omu.network.websocket_connection import WebsocketsConnection
from omu.plugin import Plugin
from omu.token import TokenProvider

from omuserver.server import Server
from omuserver.session import Session

from .plugin_connection import PluginConnection
from .plugin_session_connection import PluginSessionConnection


class PluginTokenProvider(TokenProvider):
    def __init__(self, token: str):
        self._token = token

    def get(self, server_address: Address, app: App) -> str | None:
        return self._token

    def store(self, server_address: Address, app: App, token: str) -> None:
        raise NotImplementedError


@dataclass(frozen=True, slots=True)
class PluginInstance:
    plugin: Plugin

    @classmethod
    def from_entry_point(
        cls, entry_point: importlib.metadata.EntryPoint
    ) -> PluginInstance:
        plugin = entry_point.load()
        if not isinstance(plugin, Plugin):
            raise ValueError(f"Invalid plugin: {plugin} is not a Plugin")
        return cls(plugin=plugin)

    async def start(self, server: Server):
        token = server.permission_manager.generate_plugin_token()
        pid = os.getpid()
        if self.plugin.isolated:
            process = Process(
                target=run_plugin_isolated,
                args=(
                    self.plugin,
                    server.address,
                    token,
                    pid,
                ),
                daemon=True,
            )
            process.start()
        else:
            if self.plugin.get_client is not None:
                connection = PluginConnection()
                plugin_client = self.plugin.get_client()
                plugin_client.network.set_connection(connection)
                plugin_client.network.set_token_provider(PluginTokenProvider(token))
                server.loop.create_task(plugin_client.start())
                session_connection = PluginSessionConnection(connection)
                session = await Session.from_connection(
                    server,
                    server.packet_dispatcher.packet_mapper,
                    session_connection,
                )
                server.loop.create_task(server.network.process_session(session))


def setup_logging(app: App) -> None:
    if isinstance(sys.stdout, io.TextIOWrapper):
        sys.stdout.reconfigure(encoding="utf-8")
    if isinstance(sys.stderr, io.TextIOWrapper):
        sys.stderr.reconfigure(encoding="utf-8")
    logger.add(
        f"logs/{app.id.get_sanitized_path()}/{{time:YYYY-MM-DD}}.log",
        rotation="1 day",
        colorize=False,
        format=(
            "{time:YYYY-MM-DD HH:mm:ss} | {level: <8} | "
            "{name}:{function}:{line} - {message}"
        ),
    )


def run_plugin_isolated(
    plugin: Plugin,
    address: Address,
    token: str,
    pid: int,
) -> None:
    try:
        if plugin.get_client is None:
            raise ValueError(f"Invalid plugin: {plugin} has no client")
        client = plugin.get_client()
        setup_logging(client.app)
        logger.info(f"Starting plugin {client.app.id}")
        connection = WebsocketsConnection(client, address)
        client.network.set_connection(connection)
        client.network.set_token_provider(PluginTokenProvider(token))
        loop = asyncio.new_event_loop()
        loop.set_exception_handler(asyncio_error_logger)
        client.run(loop=loop)
        loop.run_forever()
    except Exception as e:
        logger.opt(exception=e).error("Error running plugin")



from __future__ import annotations

from asyncio import Future
from collections import defaultdict
from collections.abc import Callable
from typing import TYPE_CHECKING

from loguru import logger
from omu.errors import PermissionDenied
from omu.extension.server.server_extension import (
    REQUIRE_APPS_PACKET_TYPE,
    SERVER_APP_TABLE_TYPE,
    SERVER_SESSION_TABLE_TYPE,
    SESSION_CONNECT_PACKET_TYPE,
    SESSION_DISCONNECT_PACKET_TYPE,
    SESSION_OBSERVE_PACKET_TYPE,
    SHUTDOWN_ENDPOINT_TYPE,
    TRUSTED_ORIGINS_REGISTRY_TYPE,
)
from omu.identifier import Identifier

from omuserver.helper import get_launch_command
from omuserver.server import Server
from omuserver.session import Session

from .permissions import (
    SERVER_APPS_READ_PERMISSION,
    SERVER_SESSIONS_READ_PERMISSION,
    SERVER_SHUTDOWN_PERMISSION,
    SERVER_TRUSTED_ORIGINS_GET_PERMISSION,
)

if TYPE_CHECKING:
    from loguru import Message


class WaitHandle:
    def __init__(self, ids: list[Identifier]):
        self.future = Future()
        self.ids = ids


class LogHandler:
    def __init__(
        self,
        callback: Callable[[str], None],
    ) -> None:
        self.callback = callback

    def write(self, message: Message) -> None:
        self.callback(message)


class ServerExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        server.packet_dispatcher.register(
            REQUIRE_APPS_PACKET_TYPE,
            SESSION_OBSERVE_PACKET_TYPE,
            SESSION_CONNECT_PACKET_TYPE,
            SESSION_DISCONNECT_PACKET_TYPE,
        )
        server.permission_manager.register(
            SERVER_SHUTDOWN_PERMISSION,
            SERVER_APPS_READ_PERMISSION,
            SERVER_SESSIONS_READ_PERMISSION,
            SERVER_TRUSTED_ORIGINS_GET_PERMISSION,
        )
        self.apps = self._server.tables.register(SERVER_APP_TABLE_TYPE)
        self.sessions = self._server.tables.register(SERVER_SESSION_TABLE_TYPE)
        self.trusted_origins = self._server.registries.register(
            TRUSTED_ORIGINS_REGISTRY_TYPE
        )
        server.network.event.connected += self.on_connected
        server.network.event.disconnected += self.on_disconnected
        server.event.start += self.on_start
        server.endpoints.bind_endpoint(
            SHUTDOWN_ENDPOINT_TYPE,
            self.handle_shutdown,
        )
        server.packet_dispatcher.add_packet_handler(
            REQUIRE_APPS_PACKET_TYPE, self.handle_require_apps
        )
        server.packet_dispatcher.add_packet_handler(
            SESSION_OBSERVE_PACKET_TYPE, self.handle_observe_session
        )
        self._app_waiters: dict[Identifier, list[WaitHandle]] = defaultdict(list)
        self._session_observers: dict[Identifier, list[Session]] = defaultdict(list)

    async def handle_require_apps(
        self, session: Session, app_ids: list[Identifier]
    ) -> None:
        for identifier in self._server.network._sessions.keys():
            if identifier not in app_ids:
                continue
            app_ids.remove(identifier)
        if len(app_ids) == 0:
            return

        async def task():
            waiter = WaitHandle(app_ids)
            for app_id in app_ids:
                self._app_waiters[app_id].append(waiter)
            await waiter.future

        session.add_ready_task(task)

    async def handle_observe_session(
        self, session: Session, app_ids: list[Identifier]
    ) -> None:
        if not session.permission_handle.has(SERVER_SESSIONS_READ_PERMISSION.id):
            error = f"Pemission {SERVER_SESSIONS_READ_PERMISSION.id} required to observe session"
            raise PermissionDenied(error)
        for app_id in app_ids:
            self._session_observers[app_id].append(session)

            def on_disconnect(session, app_id=app_id):
                if session in self._session_observers[app_id]:
                    self._session_observers[app_id].remove(session)

            session.event.disconnected.listen(on_disconnect)

    async def handle_shutdown(self, session: Session, restart: bool = False) -> bool:
        await self._server.shutdown()
        self._server.loop.create_task(self.shutdown(restart))
        return True

    async def shutdown(self, restart: bool = False) -> None:
        if restart:
            import os
            import sys

            os.execv(sys.executable, get_launch_command()["args"])
        else:
            self._server.loop.stop()

    async def on_start(self) -> None:
        await self.sessions.clear()

    async def on_connected(self, session: Session) -> None:
        logger.info(f"Connected: {session.app.key()}")
        await self.sessions.add(session.app)
        await self.apps.add(session.app)

        unlisten = session.event.ready.listen(self.on_session_ready)
        session.event.disconnected.listen(lambda _: unlisten())

    async def on_session_ready(self, session: Session) -> None:
        for waiter in self._app_waiters.get(session.app.id, []):
            if session.app.id in waiter.ids:
                waiter.ids.remove(session.app.id)
            if len(waiter.ids) == 0:
                waiter.future.set_result(True)
        for observer in self._session_observers.get(session.app.id, []):
            await observer.send(
                SESSION_CONNECT_PACKET_TYPE,
                session.app,
            )

    async def on_disconnected(self, session: Session) -> None:
        logger.info(f"Disconnected: {session.app.key()}")
        await self.sessions.remove(session.app)

        for observer in self._session_observers.get(session.app.id, []):
            await observer.send(
                SESSION_DISCONNECT_PACKET_TYPE,
                session.app,
            )



from __future__ import annotations

import socket

from aiohttp import web
from loguru import logger
from omu import App, Identifier
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.network.packet import PACKET_TYPES, PacketType
from omu.network.packet.packet_types import DisconnectType

from omuserver.helper import find_processes_by_port
from omuserver.network.packet_dispatcher import ServerPacketDispatcher
from omuserver.server import Server
from omuserver.session import Session, SessionType
from omuserver.session.aiohttp_connection import WebsocketsConnection


class Network:
    def __init__(
        self, server: Server, packet_dispatcher: ServerPacketDispatcher
    ) -> None:
        self._server = server
        self._packet_dispatcher = packet_dispatcher
        self._event = NetworkEvents()
        self._sessions: dict[Identifier, Session] = {}
        self._app = web.Application()
        self.add_websocket_route("/ws")
        self.register_packet(PACKET_TYPES.CONNECT, PACKET_TYPES.READY)
        self.add_packet_handler(PACKET_TYPES.READY, self._handle_ready)
        self.event.connected += self._packet_dispatcher.process_connection

    async def _handle_ready(self, session: Session, packet: None) -> None:
        await session.process_ready_tasks()
        if session.closed:
            return
        await session.send(PACKET_TYPES.READY, None)
        logger.info(f"Ready: {session.app.key()}")

    def register_packet(self, *packet_types: PacketType) -> None:
        self._packet_dispatcher.register(*packet_types)

    def add_packet_handler[T](
        self,
        packet_type: PacketType[T],
        coro: Coro[[Session, T], None],
    ) -> None:
        self._packet_dispatcher.add_packet_handler(packet_type, coro)

    def add_http_route(
        self, path: str, handle: Coro[[web.Request], web.StreamResponse]
    ) -> None:
        self._app.router.add_get(path, handle)

    async def _verify_origin(self, request: web.Request, session: Session) -> None:
        origin = request.headers.get("Origin")
        if origin is None:
            return
        namespace = session.app.id.namespace
        origin_namespace = Identifier.namespace_from_url(origin)
        if origin_namespace == namespace:
            return
        if origin_namespace in self._server.config.extra_trusted_origins:
            return
        trusted_origins = await self._server.server.trusted_origins.get()
        if origin_namespace in trusted_origins:
            return

        await session.disconnect(
            DisconnectType.INVALID_ORIGIN,
            f"Invalid origin: {origin_namespace} != {namespace}",
        )
        raise ValueError(f"Invalid origin: {origin_namespace} != {namespace}")

    def add_websocket_route(self, path: str) -> None:
        async def websocket_handler(request: web.Request) -> web.Response:
            if request.remote != "127.0.0.1":
                return web.Response(status=403)
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            connection = WebsocketsConnection(ws)
            session = await Session.from_connection(
                self._server,
                self._packet_dispatcher.packet_mapper,
                connection,
            )
            if session.kind == SessionType.APP:
                await self._verify_origin(request, session)
            await self.process_session(session)
            return ws

        self._app.router.add_get(path, websocket_handler)

    async def process_session(self, session: Session) -> None:
        if self.is_connected(session.app):
            logger.warning(f"Session {session.app} already connected")
            old_session = self._sessions[session.app.id]
            await old_session.disconnect(
                DisconnectType.ANOTHER_CONNECTION,
                f"Another connection from {session.app}",
            )
        self._sessions[session.app.id] = session
        session.event.disconnected += self.handle_disconnection
        await self._event.connected.emit(session)
        await session.listen()

    def is_connected(self, app: App) -> bool:
        return app.id in self._sessions

    async def handle_disconnection(self, session: Session) -> None:
        if session.app.id not in self._sessions:
            return
        del self._sessions[session.app.id]
        await self._event.disconnected.emit(session)

    def is_port_free(self) -> bool:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(
                    (
                        self._server.address.host or "127.0.0.1",
                        self._server.address.port,
                    )
                )
                return True
        except OSError:
            return False

    def ensure_port_availability(self):
        if not self.is_port_free():
            found_processes = set(find_processes_by_port(self._server.address.port))
            if len(found_processes) == 0:
                raise OSError(
                    f"Port {self._server.address.port} already in use by unknown process"
                )
            is_system_idle_reserved = any(
                p.pid == 0 and p.name() == "System Idle Process"
                for p in found_processes
            )
            if is_system_idle_reserved:
                raise OSError(
                    f"Port {self._server.address.port} already in use by System Idle Process"
                )
            if len(found_processes) > 1:
                processes = " ".join(f"{p.name()} ({p.pid=})" for p in found_processes)
                raise OSError(
                    f"Port {self._server.address.port} already in use by multiple processes: {processes}"
                )
            process = found_processes.pop()
            port = self._server.address.port
            name = process.name()
            pid = process.pid
            parents = process.parents()
            msg = f"Port {port} already in use by {' -> '.join(f'{p.name()}({p.pid})' for p in parents)} -> {name} ({pid=})"
            raise OSError(msg)

    async def start(self) -> None:
        self.ensure_port_availability()
        runner = web.AppRunner(self._app)
        await runner.setup()
        site = web.TCPSite(
            runner,
            host=self._server.address.host or "127.0.0.1",
            port=self._server.address.port,
        )
        await site.start()
        await self._event.start.emit()

    @property
    def event(self) -> NetworkEvents:
        return self._event


class NetworkEvents:
    def __init__(self) -> None:
        self.start = EventEmitter[[]]()
        self.connected = EventEmitter[Session]()
        self.disconnected = EventEmitter[Session]()



from dataclasses import dataclass, field

from omu.address import Address

from omuserver.directories import Directories


@dataclass(slots=True)
class Config:
    address: Address = Address(
        host="0.0.0.0",
        port=26423,
        secure=False,
    )
    debug: bool = False
    extra_trusted_origins: list[str] = field(default_factory=list)
    directories: Directories = field(default_factory=Directories.default)
    dashboard_token: str | None = None



import pathlib
from dataclasses import dataclass
from typing import LiteralString


@dataclass(frozen=True, slots=True)
class Directories:
    data: pathlib.Path
    assets: pathlib.Path
    index: pathlib.Path
    version: pathlib.Path

    @classmethod
    def default(cls):
        cwd = pathlib.Path.cwd()
        return Directories(
            data=cwd / "data",
            assets=cwd / "assets",
            index=pathlib.Path(__file__).parent / "index.html",
            version=cwd / "VERSION",
        )

    def mkdir(self):
        self.data.mkdir(parents=True, exist_ok=True)
        self.assets.mkdir(parents=True, exist_ok=True)

    def get(self, name: LiteralString):
        path = self.data / name
        path.mkdir(parents=True, exist_ok=True)
        return path



from __future__ import annotations

import asyncio
import importlib.metadata
import importlib.util
import sys
import tempfile
from collections.abc import Mapping
from typing import (
    Protocol,
)

import aiohttp
import uv
from loguru import logger
from omu.extension.plugin import PackageInfo, PluginPackageInfo
from omu.plugin import Plugin
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from omuserver.server import Server

from .plugin_instance import PluginInstance

PLUGIN_GROUP = "omu.plugins"


class PluginModule(Protocol):
    plugin: Plugin


class DependencyResolver:
    def __init__(self) -> None:
        self._dependencies: dict[str, SpecifierSet] = {}
        self._packages_distributions: Mapping[str, importlib.metadata.Distribution] = {}
        self._packages_distributions_changed = True
        self.find_packages_distributions()

    async def fetch_package_info(self, package: str) -> PackageInfo:
        async with aiohttp.ClientSession() as session:
            async with session.get(f"https://pypi.org/pypi/{package}/json") as response:
                return await response.json()

    async def get_installed_package_info(
        self, package: str
    ) -> PluginPackageInfo | None:
        try:
            package_info = importlib.metadata.distribution(package)
        except importlib.metadata.PackageNotFoundError:
            return None
        return PluginPackageInfo(
            package=package_info.name,
            version=package_info.version,
        )

    def format_dependencies(
        self, dependencies: Mapping[str, SpecifierSet | None]
    ) -> list[str]:
        args = []
        for dependency, specifier in dependencies.items():
            if specifier is not None:
                args.append(f"{dependency}{specifier}")
            else:
                args.append(dependency)
        return args

    async def update_requirements(self, requirements: dict[str, SpecifierSet]) -> None:
        if len(requirements) == 0:
            return
        with tempfile.NamedTemporaryFile(mode="wb", delete=True) as req_file:
            dependency_lines = self.format_dependencies(requirements)
            req_file.write("\n".join(dependency_lines).encode("utf-8"))
            req_file.flush()
            process = await asyncio.create_subprocess_exec(
                uv.find_uv_bin(),
                "pip",
                "install",
                "--upgrade",
                "-r",
                req_file.name,
                "--python",
                sys.executable,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await process.communicate()
        if process.returncode != 0:
            logger.error(f"Error running uv command: {stderr}")
            return
        logger.info(f"Ran uv command: {(stdout or stderr).decode()}")

    def is_package_satisfied(self, package: str, specifier: SpecifierSet) -> bool:
        package_info = self._packages_distributions.get(package)
        if package_info is None:
            return False
        installed_version = Version(package_info.version)
        return installed_version in specifier

    def add_dependencies(self, dependencies: Mapping[str, SpecifierSet | None]) -> bool:
        changed = False
        for dependency, specifier in dependencies.items():
            if dependency not in self._dependencies:
                if specifier is not None and self.is_package_satisfied(
                    dependency, specifier
                ):
                    continue
                self._dependencies[dependency] = specifier or SpecifierSet()
                changed = True
                continue
            if specifier is not None:
                specifier_set = self._dependencies[dependency]
                if specifier_set != specifier:
                    changed = True
                specifier_set &= specifier
                continue
        return changed

    def find_packages_distributions(
        self,
    ) -> Mapping[str, importlib.metadata.Distribution]:
        if not self._packages_distributions_changed:
            return self._packages_distributions
        self._packages_distributions: Mapping[str, importlib.metadata.Distribution] = {
            dist.name: dist for dist in importlib.metadata.distributions()
        }
        self._packages_distributions_changed = False
        return self._packages_distributions

    async def resolve(self):
        packages_distributions = self.find_packages_distributions()
        requirements: dict[str, SpecifierSet] = {}
        skipped: dict[str, SpecifierSet] = {}
        for dependency, specifier in self._dependencies.items():
            package = packages_distributions.get(dependency)
            if package is None:
                requirements[dependency] = specifier
                continue
            distribution = packages_distributions[package.name]
            installed_version = Version(distribution.version)
            specifier_set = self._dependencies[dependency]
            if installed_version in specifier_set:
                skipped[dependency] = specifier_set
                continue
            requirements[dependency] = specifier_set
        if len(requirements) == 0:
            return

        await self.update_requirements(requirements)
        self._packages_distributions_changed = True


class PluginLoader:
    def __init__(self, server: Server) -> None:
        self._server = server
        self.instances: dict[str, PluginInstance] = {}
        server.event.stop += self.handle_server_stop

    async def handle_server_stop(self) -> None:
        for instance in self.instances.values():
            if instance.plugin.on_stop_server is not None:
                await instance.plugin.on_stop_server(self._server)

    async def run_plugins(self):
        self.load_plugins_from_entry_points()

        logger.info(f"Loaded plugins: {self.instances.keys()}")

        for instance in self.instances.values():
            await self.start_plugin(instance)

    def load_plugins_from_entry_points(self):
        entry_points = importlib.metadata.entry_points(group=PLUGIN_GROUP)
        for entry_point in entry_points:
            if entry_point.dist is None:
                raise ValueError(f"Invalid plugin: {entry_point} has no distribution")
            plugin_key = entry_point.dist.name
            if plugin_key in self.instances:
                raise ValueError(f"Duplicate plugin: {entry_point}")
            plugin = PluginInstance.from_entry_point(entry_point)
            self.instances[plugin_key] = plugin

    async def load_updated_plugins(self):
        entry_points = importlib.metadata.entry_points(group=PLUGIN_GROUP)
        detected_plugins = {
            entry_point.dist.name for entry_point in entry_points if entry_point.dist
        }
        logger.info(f"Detected plugins: {detected_plugins}")
        for entry_point in entry_points:
            if entry_point.dist is None:
                raise ValueError(f"Invalid plugin: {entry_point} has no distribution")
            plugin_key = entry_point.dist.name
            if plugin_key in self.instances:
                continue
            instance = PluginInstance.from_entry_point(entry_point)
            self.instances[plugin_key] = instance
            await self.start_plugin(instance)

    async def start_plugin(self, instance: PluginInstance):
        try:
            if instance.plugin.on_start_server is not None:
                await instance.plugin.on_start_server(self._server)

            await instance.start(self._server)
        except Exception as e:
            logger.opt(exception=e).error(f"Error starting plugin: {instance.plugin}")



from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, Mapping

from omu.extension.table import TableConfig, TablePermissions
from omu.extension.table.table_extension import TABLE_PROXY_PACKET, TableProxyPacket
from omu.identifier import Identifier

from omuserver.server import Server
from omuserver.session import Session

from .adapters.tableadapter import TableAdapter
from .server_table import ServerTable, ServerTableEvents
from .session_table_handler import SessionTableListener


class CachedTable(ServerTable):
    def __init__(
        self,
        server: Server,
        id: Identifier,
    ):
        self._server = server
        self._id = id
        self._event = ServerTableEvents()
        self._sessions: dict[Session, SessionTableListener] = {}
        self._permissions: TablePermissions | None = None
        self._proxy_sessions: dict[str, Session] = {}
        self._changed = False
        self._proxy_id = 0
        self._save_task: asyncio.Task | None = None
        self._adapter: TableAdapter | None = None
        self.config: TableConfig = {}
        self._cache: dict[str, bytes] = {}
        self._cache_size: int | None = None

    def set_config(self, config: TableConfig) -> None:
        self.config = config
        self._cache_size = config.get("cache_size", None)

    @property
    def permissions(self) -> TablePermissions | None:
        return self._permissions

    def set_permissions(self, permissions: TablePermissions | None) -> None:
        self._permissions = permissions

    def set_adapter(self, adapter: TableAdapter) -> None:
        self._adapter = adapter

    async def load(self) -> None:
        if self._adapter is None:
            raise Exception("Table not set")
        await self._adapter.load()

    async def store(self) -> None:
        if self._adapter is None:
            raise Exception("Table not set")
        if not self._changed:
            return
        await self._adapter.store()
        self._changed = False

    def attach_session(self, session: Session) -> None:
        if session in self._sessions:
            return
        handler = SessionTableListener(
            id=self._id,
            session=session,
            table=self,
        )
        self._sessions[session] = handler
        session.event.disconnected += self.handle_disconnection

    def detach_session(self, session: Session) -> None:
        if session in self._proxy_sessions:
            del self._proxy_sessions[session.app.key()]
        if session in self._sessions:
            handler = self._sessions.pop(session)
            handler.close()

    async def handle_disconnection(self, session: Session) -> None:
        self.detach_session(session)

    def attach_proxy_session(self, session: Session) -> None:
        self._proxy_sessions[session.app.key()] = session

    async def get(self, key: str) -> bytes | None:
        if self._adapter is None:
            raise Exception("Table not set")
        if key in self._cache:
            return self._cache[key]
        data = await self._adapter.get(key)
        if data is None:
            return None
        await self.update_cache({key: data})
        return data

    async def get_many(self, *keys: str) -> dict[str, bytes]:
        key_list = list(keys)
        if self._adapter is None:
            raise Exception("Table not set")
        items: dict[str, bytes] = {}
        for key in tuple(key_list):
            if key in self._cache:
                items[key] = self._cache[key]
                key_list.remove(key)
        if len(key_list) == 0:
            return items
        data = await self._adapter.get_many(key_list)
        for key, value in data.items():
            items[key] = value
        await self.update_cache(items)
        return items

    async def add(self, items: Mapping[str, bytes]) -> None:
        if self._adapter is None:
            raise Exception("Table not set")
        if len(self._proxy_sessions) > 0:
            await self.send_proxy_event(items)
            return
        await self._adapter.set_all(items)
        await self._event.add(items)
        await self.update_cache(items)
        self.mark_changed()

    async def send_proxy_event(self, items: Mapping[str, bytes]) -> None:
        session = tuple(self._proxy_sessions.values())[0]
        self._proxy_id += 1
        await session.send(
            TABLE_PROXY_PACKET,
            TableProxyPacket(
                id=self._id,
                items=items,
                key=self._proxy_id,
            ),
        )

    async def proxy(
        self, session: Session, key: int, items: Mapping[str, bytes]
    ) -> int:
        adapter = self._adapter
        if adapter is None:
            raise Exception("Table not set")
        if session.app.key() not in self._proxy_sessions:
            raise ValueError("Session not in proxy sessions")
        session_key = session.app.key()
        index = tuple(self._proxy_sessions.keys()).index(session_key)
        if index == len(self._proxy_sessions) - 1:
            adapter = self._adapter
            if adapter is None:
                raise Exception("Table not set")
            await adapter.set_all(items)
            await self._event.add(items)
            await self.update_cache(items)
            self.mark_changed()
            return 0
        session = tuple(self._proxy_sessions.values())[index + 1]
        await session.send(
            TABLE_PROXY_PACKET,
            TableProxyPacket(
                id=self._id,
                items=items,
                key=self._proxy_id,
            ),
        )
        return self._proxy_id

    async def update(self, items: Mapping[str, bytes]) -> None:
        if self._adapter is None:
            raise Exception("Table not set")
        await self._adapter.set_all(items)
        await self._event.update(items)
        await self.update_cache(items)
        self.mark_changed()

    async def remove(self, keys: list[str] | tuple[str, ...]) -> None:
        if self._adapter is None:
            raise Exception("Table not set")
        removed = await self._adapter.get_many(keys)
        await self._adapter.remove_all(keys)
        for key in keys:
            if key in self._cache:
                del self._cache[key]
        await self._event.remove(removed)
        self.mark_changed()

    async def clear(self) -> None:
        if self._adapter is None:
            raise Exception("Table not set")
        await self._adapter.clear()
        await self._event.clear()
        self._cache.clear()
        self.mark_changed()

    async def has(self, key: str) -> bool:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.has(key)

    async def has_many(self, keys: list[str] | tuple[str, ...]) -> dict[str, bool]:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.has_many(list(keys))

    async def has_all(self, keys: list[str] | tuple[str, ...]) -> bool:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.has_all(list(keys))

    async def has_any(self, keys: list[str] | tuple[str, ...]) -> bool:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.has_any(list(keys))

    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> dict[str, bytes]:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.fetch_items(limit, backward, cursor)

    async def fetch_range(self, start: str, end: str) -> dict[str, bytes]:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.fetch_range(start, end)

    async def fetch_all(self) -> dict[str, bytes]:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.fetch_all()

    async def iterate(self) -> AsyncGenerator[bytes, None]:
        cursor: str | None = None
        while True:
            items = await self.fetch_items(
                limit=self.config.get("chunk_size", 100),
                cursor=cursor,
            )
            if len(items) == 0:
                break
            for item in items.values():
                yield item
            *_, cursor = items.keys()

    async def size(self) -> int:
        if self._adapter is None:
            raise Exception("Table not set")
        return await self._adapter.size()

    async def save_task(self) -> None:
        while self._changed:
            await self.store()
            await asyncio.sleep(30)

    def mark_changed(self) -> None:
        self._changed = True
        if self._save_task is None:
            self._save_task = asyncio.create_task(self.save_task())

    def set_cache_size(self, size: int) -> None:
        self._cache_size = size

    async def update_cache(self, items: Mapping[str, bytes]) -> None:
        if self._cache_size is None or self._cache_size <= 0:
            return
        for key, item in items.items():
            self._cache[key] = item
            if len(self._cache) > self._cache_size:
                del self._cache[next(iter(self._cache))]
        await self._event.cache_update(self._cache)

    @property
    def cache(self) -> Mapping[str, bytes]:
        return self._cache

    @property
    def event(self) -> ServerTableEvents:
        return self._event

    @property
    def id(self) -> Identifier:
        return self._id

    @property
    def adapter(self) -> TableAdapter | None:
        return self._adapter



import json
import time
from asyncio import Future

from omu.app import App
from omu.errors import PermissionDenied
from omu.extension.dashboard.dashboard_extension import (
    DASHBOARD_APP_INSTALL_ACCEPT_PACKET,
    DASHBOARD_APP_INSTALL_DENY_PACKET,
    DASHBOARD_APP_INSTALL_ENDPOINT,
    DASHBOARD_APP_INSTALL_PACKET,
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    DASHBOARD_APP_TABLE_TYPE,
    DASHBOARD_APP_UPDATE_ACCEPT_PACKET,
    DASHBOARD_APP_UPDATE_DENY_PACKET,
    DASHBOARD_APP_UPDATE_ENDPOINT,
    DASHBOARD_APP_UPDATE_PACKET,
    DASHBOARD_APP_UPDATE_PERMISSION_ID,
    DASHBOARD_OPEN_APP_ENDPOINT,
    DASHBOARD_OPEN_APP_PACKET,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_PERMISSION_ACCEPT_PACKET,
    DASHBOARD_PERMISSION_DENY_PACKET,
    DASHBOARD_PERMISSION_REQUEST_PACKET,
    DASHBOARD_PLUGIN_ACCEPT_PACKET,
    DASHBOARD_PLUGIN_DENY_PACKET,
    DASHBOARD_PLUGIN_REQUEST_PACKET,
    DASHBOARD_SET_ENDPOINT,
    DashboardSetResponse,
)
from omu.extension.dashboard.packets import (
    AppInstallRequestPacket,
    AppInstallResponse,
    AppUpdateRequestPacket,
    AppUpdateResponse,
    PermissionRequestPacket,
    PluginRequestPacket,
)
from omu.identifier import Identifier

from omuserver.server import Server
from omuserver.session import Session
from omuserver.session.session import SessionType

from .permission import (
    DASHBOARD_APP_INSTALL_PERMISSION,
    DASHBOARD_APP_UPDATE_PERMISSION,
    DASHBOARD_OPEN_APP_PERMISSION,
    DASHBOARD_SET_PERMISSION,
    DASHOBARD_APP_EDIT_PERMISSION,
    DASHOBARD_APP_READ_PERMISSION,
)


class DashboardExtension:
    def __init__(self, server: Server) -> None:
        server.packet_dispatcher.register(
            DASHBOARD_PERMISSION_REQUEST_PACKET,
            DASHBOARD_PERMISSION_ACCEPT_PACKET,
            DASHBOARD_PERMISSION_DENY_PACKET,
            DASHBOARD_PLUGIN_REQUEST_PACKET,
            DASHBOARD_PLUGIN_ACCEPT_PACKET,
            DASHBOARD_PLUGIN_DENY_PACKET,
            DASHBOARD_APP_INSTALL_PACKET,
            DASHBOARD_APP_INSTALL_ACCEPT_PACKET,
            DASHBOARD_APP_INSTALL_DENY_PACKET,
            DASHBOARD_APP_UPDATE_PACKET,
            DASHBOARD_APP_UPDATE_ACCEPT_PACKET,
            DASHBOARD_APP_UPDATE_DENY_PACKET,
        )
        server.permission_manager.register(
            DASHBOARD_SET_PERMISSION,
            DASHBOARD_OPEN_APP_PERMISSION,
            DASHOBARD_APP_READ_PERMISSION,
            DASHOBARD_APP_EDIT_PERMISSION,
            DASHBOARD_APP_INSTALL_PERMISSION,
            DASHBOARD_APP_UPDATE_PERMISSION,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_PERMISSION_ACCEPT_PACKET,
            self.handle_permission_accept,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_PERMISSION_DENY_PACKET,
            self.handle_permission_deny,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_PLUGIN_ACCEPT_PACKET,
            self.handle_plugin_accept,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_PLUGIN_DENY_PACKET,
            self.handle_plugin_deny,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_APP_INSTALL_ACCEPT_PACKET,
            self.handle_app_install_accept,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_APP_INSTALL_DENY_PACKET,
            self.handle_app_install_deny,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_APP_UPDATE_ACCEPT_PACKET,
            self.handle_app_update_accept,
        )
        server.packet_dispatcher.add_packet_handler(
            DASHBOARD_APP_UPDATE_DENY_PACKET,
            self.handle_app_update_deny,
        )

        server.endpoints.bind_endpoint(
            DASHBOARD_SET_ENDPOINT,
            self.handle_dashboard_set,
        )
        server.endpoints.bind_endpoint(
            DASHBOARD_OPEN_APP_ENDPOINT,
            self.handle_dashboard_open_app,
        )
        server.endpoints.bind_endpoint(
            DASHBOARD_APP_INSTALL_ENDPOINT,
            self.handle_dashboard_app_install,
        )
        server.endpoints.bind_endpoint(
            DASHBOARD_APP_UPDATE_ENDPOINT,
            self.handle_dashboard_app_update,
        )
        server.network.event.connected += self.handle_session_connected
        self.server = server
        self.apps = server.tables.register(DASHBOARD_APP_TABLE_TYPE)
        self.dashboard_session: Session | None = None
        self.pending_permission_requests: dict[str, PermissionRequestPacket] = {}
        self.permission_requests: dict[str, Future[bool]] = {}
        self.pending_plugin_requests: dict[str, PluginRequestPacket] = {}
        self.plugin_requests: dict[str, Future[bool]] = {}
        self.pending_app_install_requests: dict[str, AppInstallRequestPacket] = {}
        self.app_install_requests: dict[str, Future[bool]] = {}
        self.pending_app_update_requests: dict[str, AppUpdateRequestPacket] = {}
        self.app_update_requests: dict[str, Future[bool]] = {}
        self.request_id = 0

    async def handle_session_connected(self, session: Session) -> None:
        if session.kind != SessionType.APP:
            return
        exist_app = await self.apps.get(session.app.id.key())
        if exist_app is None:
            return
        old_metadata = json.dumps(exist_app.metadata)
        new_metadata = json.dumps(session.app.metadata)
        if old_metadata != new_metadata:
            await self.update_app(session.app)

    async def handle_dashboard_open_app(self, session: Session, app: App) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        if not session.permission_handle.has(DASHBOARD_OPEN_APP_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to open apps")
        await self.dashboard_session.send(
            DASHBOARD_OPEN_APP_PACKET,
            app,
        )

    async def handle_dashboard_set(
        self, session: Session, identifier: Identifier
    ) -> DashboardSetResponse:
        if session.kind != SessionType.DASHBOARD:
            raise PermissionDenied("Session is not a dashboard")
        self.dashboard_session = session
        session.event.disconnected += self._on_dashboard_disconnected
        await self.send_pending_permission_requests()
        await self.send_pending_plugin_requests()
        await self.send_pending_app_install_requests()
        await self.send_pending_app_update_requests()
        return {"success": True}

    async def send_pending_permission_requests(self) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        for request in self.pending_permission_requests.values():
            await self.dashboard_session.send(
                DASHBOARD_PERMISSION_REQUEST_PACKET,
                request,
            )

    async def send_pending_plugin_requests(self) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        for request in self.pending_plugin_requests.values():
            await self.dashboard_session.send(
                DASHBOARD_PLUGIN_REQUEST_PACKET,
                request,
            )

    async def send_pending_app_install_requests(self) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        for request in self.pending_app_install_requests.values():
            await self.dashboard_session.send(
                DASHBOARD_APP_INSTALL_PACKET,
                request,
            )

    async def send_pending_app_update_requests(self) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        for request in self.pending_app_update_requests.values():
            await self.dashboard_session.send(
                DASHBOARD_APP_UPDATE_PACKET,
                request,
            )

    async def _on_dashboard_disconnected(self, session: Session) -> None:
        self.dashboard_session = None

    def ensure_dashboard_session(self, session: Session) -> bool:
        if session == self.dashboard_session:
            return True
        msg = f"Session {session} is not the dashboard session"
        raise PermissionDenied(msg)

    async def handle_permission_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.permission_requests:
            raise ValueError(f"Permission request with id {request_id} does not exist")
        del self.pending_permission_requests[request_id]
        future = self.permission_requests.pop(request_id)
        future.set_result(True)

    async def handle_permission_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.permission_requests:
            raise ValueError(f"Permission request with id {request_id} does not exist")
        del self.pending_permission_requests[request_id]
        future = self.permission_requests.pop(request_id)
        future.set_result(False)

    async def request_permissions(self, request: PermissionRequestPacket) -> bool:
        if request.request_id in self.permission_requests:
            raise ValueError(
                f"Permission request with id {request.request_id} already exists"
            )
        future = Future[bool]()
        self.permission_requests[request.request_id] = future
        await self.notify_dashboard_permission_request(request)
        return await future

    async def notify_dashboard_permission_request(
        self, request: PermissionRequestPacket
    ) -> None:
        self.pending_permission_requests[request.request_id] = request
        if self.dashboard_session is not None:
            await self.dashboard_session.send(
                DASHBOARD_PERMISSION_REQUEST_PACKET,
                request,
            )

    async def handle_plugin_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.plugin_requests:
            raise ValueError(f"Plugin request with id {request_id} does not exist")
        del self.pending_plugin_requests[request_id]
        future = self.plugin_requests.pop(request_id)
        future.set_result(True)

    async def handle_plugin_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.plugin_requests:
            raise ValueError(f"Plugin request with id {request_id} does not exist")
        del self.pending_plugin_requests[request_id]
        future = self.plugin_requests.pop(request_id)
        future.set_result(False)

    async def request_plugins(self, request: PluginRequestPacket) -> bool:
        if request.request_id in self.plugin_requests:
            raise ValueError(
                f"Plugin request with id {request.request_id} already exists"
            )
        future = Future[bool]()
        self.plugin_requests[request.request_id] = future
        await self.notify_dashboard_plugin_request(request)
        return await future

    async def notify_dashboard_plugin_request(
        self, request: PluginRequestPacket
    ) -> None:
        self.pending_plugin_requests[request.request_id] = request
        if self.dashboard_session is not None:
            await self.dashboard_session.send(
                DASHBOARD_PLUGIN_REQUEST_PACKET,
                request,
            )

    async def handle_app_install_accept(
        self, session: Session, request_id: str
    ) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_install_requests:
            raise ValueError(f"App install request with id {request_id} does not exist")
        del self.pending_app_install_requests[request_id]
        future = self.app_install_requests.pop(request_id)
        future.set_result(True)

    async def handle_app_install_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_install_requests:
            raise ValueError(f"App install request with id {request_id} does not exist")
        del self.pending_app_install_requests[request_id]
        future = self.app_install_requests.pop(request_id)
        future.set_result(False)

    async def handle_dashboard_app_install(
        self, session: Session, app: App
    ) -> AppInstallResponse:
        if not session.permission_handle.has(DASHBOARD_APP_INSTALL_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to install apps")
        request_id = self._get_next_request_id()
        future = Future[bool]()
        self.app_install_requests[request_id] = future
        await self.notify_dashboard_app_install(
            AppInstallRequestPacket(request_id=request_id, app=app)
        )
        accepted = await future
        if accepted:
            await self.apps.add(app)
        return AppInstallResponse(accepted=accepted)

    async def notify_dashboard_app_install(
        self, request: AppInstallRequestPacket
    ) -> None:
        self.pending_app_install_requests[request.request_id] = request
        if self.dashboard_session is not None:
            await self.dashboard_session.send(
                DASHBOARD_APP_INSTALL_PACKET,
                request,
            )

    async def handle_app_update_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_update_requests:
            raise ValueError(f"App update request with id {request_id} does not exist")
        del self.pending_app_update_requests[request_id]
        future = self.app_update_requests.pop(request_id)
        future.set_result(True)

    async def handle_app_update_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_update_requests:
            raise ValueError(f"App update request with id {request_id} does not exist")
        del self.pending_app_update_requests[request_id]
        future = self.app_update_requests.pop(request_id)
        future.set_result(False)

    async def handle_dashboard_app_update(
        self, session: Session, app: App
    ) -> AppUpdateResponse:
        if not session.permission_handle.has(DASHBOARD_APP_UPDATE_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to update apps")
        accepted = await self.update_app(app)
        return AppUpdateResponse(accepted=accepted)

    async def update_app(self, app):
        old_app = await self.apps.get(app.id.key())
        if old_app is None:
            raise ValueError(f"App with id {app.id} does not exist")
        request_id = self._get_next_request_id()
        future = Future[bool]()
        self.app_update_requests[request_id] = future
        await self.notify_dashboard_app_update(
            AppUpdateRequestPacket(
                request_id=request_id,
                old_app=old_app,
                new_app=app,
            )
        )
        accepted = await future
        if accepted:
            await self.apps.add(app)
        return accepted

    async def notify_dashboard_app_update(
        self, request: AppUpdateRequestPacket
    ) -> None:
        self.pending_app_update_requests[request.request_id] = request
        if self.dashboard_session is not None:
            await self.dashboard_session.send(
                DASHBOARD_APP_UPDATE_PACKET,
                request,
            )

    def _get_next_request_id(self) -> str:
        self.request_id += 1
        return f"{self.request_id}-{time.time_ns()}"



from omu.extension.i18n.i18n_extension import I18N_LOCALES_REGISTRY_TYPE

from omuserver.server import Server

from .permissions import (
    I18N_GET_LOCALES_PERMISSION,
    I18N_SET_LOCALES_PERMISSION,
)


class I18nExtension:
    def __init__(self, server: Server):
        server.permission_manager.register(
            I18N_GET_LOCALES_PERMISSION,
            I18N_SET_LOCALES_PERMISSION,
        )
        server.registries.register(I18N_LOCALES_REGISTRY_TYPE)



from __future__ import annotations

import abc
from pathlib import Path

from omu.extension.asset.asset_extension import (
    ASSET_DOWNLOAD_ENDPOINT,
    ASSET_DOWNLOAD_MANY_ENDPOINT,
    ASSET_UPLOAD_ENDPOINT,
    ASSET_UPLOAD_MANY_ENDPOINT,
    File,
)
from omu.identifier import Identifier

from omuserver.helper import safe_path_join
from omuserver.server import Server
from omuserver.session import Session

from .permissions import (
    ASSET_DOWNLOAD_PERMISSION,
    ASSET_UPLOAD_PERMISSION,
)


class AssetStorage(abc.ABC):
    @abc.abstractmethod
    async def store(self, file: File) -> Identifier: ...

    @abc.abstractmethod
    async def retrieve(self, identifier: Identifier) -> File: ...


class FileStorage(AssetStorage):
    def __init__(self, path: Path) -> None:
        self._path = path

    async def store(self, file: File) -> Identifier:
        path = file.identifier.get_sanitized_path()
        file_path = safe_path_join(self._path, path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_bytes(file.buffer)
        return file.identifier

    async def retrieve(self, identifier: Identifier) -> File:
        path = identifier.get_sanitized_path()
        file_path = safe_path_join(self._path, path)
        return File(identifier, file_path.read_bytes())


class AssetExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        self.storage = FileStorage(server.directories.assets)
        server.permission_manager.register(
            ASSET_UPLOAD_PERMISSION,
            ASSET_DOWNLOAD_PERMISSION,
        )
        server.endpoints.bind_endpoint(
            ASSET_UPLOAD_ENDPOINT,
            self.handle_upload,
        )
        server.endpoints.bind_endpoint(
            ASSET_UPLOAD_MANY_ENDPOINT,
            self.handle_upload_many,
        )
        server.endpoints.bind_endpoint(
            ASSET_DOWNLOAD_ENDPOINT,
            self.handle_download,
        )
        server.endpoints.bind_endpoint(
            ASSET_DOWNLOAD_MANY_ENDPOINT,
            self.handle_download_many,
        )

    async def handle_upload(self, session: Session, file: File) -> Identifier:
        identifier = await self.storage.store(file)
        return identifier

    async def handle_upload_many(
        self, session: Session, files: list[File]
    ) -> list[Identifier]:
        identifiers: list[Identifier] = []
        for file in files:
            identifier = await self.storage.store(file)
            identifiers.append(identifier)
        return identifiers

    async def handle_download(self, session: Session, identifier: Identifier) -> File:
        return await self.storage.retrieve(identifier)

    async def handle_download_many(
        self, session: Session, identifiers: list[Identifier]
    ) -> list[File]:
        files: list[File] = []
        for identifier in identifiers:
            file = await self.storage.retrieve(identifier)
            files.append(file)
        return files



from __future__ import annotations

import abc
import asyncio
from typing import TYPE_CHECKING

from omu.address import Address
from omu.event_emitter import EventEmitter

if TYPE_CHECKING:
    from omuserver.config import Config
    from omuserver.directories import Directories
    from omuserver.extension.asset import AssetExtension
    from omuserver.extension.dashboard import DashboardExtension
    from omuserver.extension.endpoint import EndpointExtension
    from omuserver.extension.i18n import I18nExtension
    from omuserver.extension.logger import LoggerExtension
    from omuserver.extension.permission import PermissionExtension
    from omuserver.extension.plugin import PluginExtension
    from omuserver.extension.registry import RegistryExtension
    from omuserver.extension.server import ServerExtension
    from omuserver.extension.signal import SignalExtension
    from omuserver.extension.table import TableExtension
    from omuserver.network import Network
    from omuserver.network.packet_dispatcher import ServerPacketDispatcher
    from omuserver.security import PermissionManager


class ServerEvents:
    def __init__(self) -> None:
        self.start = EventEmitter[[]]()
        self.stop = EventEmitter[[]]()


class Server(abc.ABC):
    @property
    @abc.abstractmethod
    def config(self) -> Config: ...

    @property
    @abc.abstractmethod
    def loop(self) -> asyncio.AbstractEventLoop: ...

    @property
    @abc.abstractmethod
    def address(self) -> Address: ...

    @property
    @abc.abstractmethod
    def directories(self) -> Directories: ...

    @property
    @abc.abstractmethod
    def permission_manager(self) -> PermissionManager: ...

    @property
    @abc.abstractmethod
    def network(self) -> Network: ...

    @property
    @abc.abstractmethod
    def packet_dispatcher(self) -> ServerPacketDispatcher: ...

    @property
    @abc.abstractmethod
    def endpoints(self) -> EndpointExtension: ...
    @property
    @abc.abstractmethod
    def permissions(self) -> PermissionExtension: ...
    @property
    @abc.abstractmethod
    def tables(self) -> TableExtension: ...
    @property
    @abc.abstractmethod
    def dashboard(self) -> DashboardExtension: ...
    @property
    @abc.abstractmethod
    def registries(self) -> RegistryExtension: ...
    @property
    @abc.abstractmethod
    def server(self) -> ServerExtension: ...
    @property
    @abc.abstractmethod
    def signals(self) -> SignalExtension: ...
    @property
    @abc.abstractmethod
    def plugins(self) -> PluginExtension: ...
    @property
    @abc.abstractmethod
    def assets(self) -> AssetExtension: ...
    @property
    @abc.abstractmethod
    def i18n(self) -> I18nExtension: ...
    @property
    @abc.abstractmethod
    def logger(self) -> LoggerExtension: ...

    @property
    @abc.abstractmethod
    def running(self) -> bool: ...

    @abc.abstractmethod
    def run(self) -> None: ...

    @abc.abstractmethod
    async def start(self) -> None: ...

    @abc.abstractmethod
    async def shutdown(self) -> None: ...

    @property
    @abc.abstractmethod
    def event(self) -> ServerEvents: ...



from omu.extension.asset import (
    ASSET_DOWNLOAD_PERMISSION_ID,
    ASSET_UPLOAD_PERMISSION_ID,
)
from omu.extension.permission import PermissionType

ASSET_UPLOAD_PERMISSION = PermissionType(
    id=ASSET_UPLOAD_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "ファイルを保存",
            "en": "Save a file",
        },
        "note": {
            "ja": "アプリがファイルを保持するために使われます",
            "en": "Used by apps to store files",
        },
    },
)
ASSET_DOWNLOAD_PERMISSION = PermissionType(
    id=ASSET_DOWNLOAD_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "ファイルをダウンロード",
            "en": "Download a file",
        },
        "note": {
            "ja": "アプリが保存したファイルを読み込みなおすために使われます",
            "en": "Used by apps to reload saved files",
        },
    },
)



import asyncio
import json

import aiohttp
from aiohttp import web
from loguru import logger
from omu import Identifier
from omu.address import Address
from omu.helper import asyncio_error_logger

from omuserver import __version__
from omuserver.config import Config
from omuserver.directories import Directories
from omuserver.extension.asset import AssetExtension
from omuserver.extension.dashboard import DashboardExtension
from omuserver.extension.endpoint import EndpointExtension
from omuserver.extension.i18n import I18nExtension
from omuserver.extension.logger import LoggerExtension
from omuserver.extension.permission import PermissionExtension
from omuserver.extension.plugin import PluginExtension
from omuserver.extension.registry import RegistryExtension
from omuserver.extension.server import ServerExtension
from omuserver.extension.signal import SignalExtension
from omuserver.extension.table import TableExtension
from omuserver.helper import safe_path_join
from omuserver.network import Network
from omuserver.network.packet_dispatcher import ServerPacketDispatcher
from omuserver.security.security import PermissionManager, ServerPermissionManager

from .server import Server, ServerEvents

USER_AGENT_HEADERS = {
    "User-Agent": json.dumps(
        [
            "omu",
            {
                "name": "omuserver",
                "version": __version__,
            },
        ]
    )
}


class OmuServer(Server):
    def __init__(
        self,
        config: Config,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        self._config = config
        self._loop = loop or self.create_loop()
        self._address = config.address
        self._event = ServerEvents()
        self._directories = config.directories
        self._directories.mkdir()
        self._packet_dispatcher = ServerPacketDispatcher()
        self._network = Network(self, self._packet_dispatcher)
        self._network.event.start += self._handle_network_start
        self._network.add_http_route("/", self._handle_index)
        self._network.add_http_route("/version", self._handle_version)
        self._network.add_http_route("/proxy", self._handle_proxy)
        self._network.add_http_route("/asset", self._handle_assets)
        self._security = ServerPermissionManager(self)
        self._running = False
        self._endpoints = EndpointExtension(self)
        self._permissions = PermissionExtension(self)
        self._tables = TableExtension(self)
        self._dashboard = DashboardExtension(self)
        self._registries = RegistryExtension(self)
        self._server = ServerExtension(self)
        self._signals = SignalExtension(self)
        self._plugins = PluginExtension(self)
        self._assets = AssetExtension(self)
        self._i18n = I18nExtension(self)
        self._logger = LoggerExtension(self)
        self.client = aiohttp.ClientSession(
            loop=self.loop,
            headers=USER_AGENT_HEADERS,
        )

    def create_loop(self) -> asyncio.AbstractEventLoop:
        loop = asyncio.new_event_loop()
        loop.set_exception_handler(asyncio_error_logger)
        return loop

    async def _handle_index(self, request: web.Request) -> web.StreamResponse:
        return web.FileResponse(self._directories.index)

    async def _handle_version(self, request: web.Request) -> web.Response:
        return web.json_response(
            {
                "version": __version__,
            }
        )

    async def _handle_proxy(self, request: web.Request) -> web.StreamResponse:
        url = request.query.get("url")
        no_cache = bool(request.query.get("no_cache"))
        if not url:
            return web.Response(status=400)
        try:
            async with self.client.get(url) as resp:
                headers = {
                    "Cache-Control": "no-cache" if no_cache else "max-age=3600",
                    "Content-Type": resp.content_type,
                    "Access-Control-Allow-Origin": "*",
                }
                response = web.StreamResponse(
                    status=resp.status,
                    headers=headers,
                )
                await response.prepare(request)
                try:
                    async for chunk in resp.content.iter_any():
                        await response.write(chunk)
                except ConnectionResetError:
                    pass
                return response
        except aiohttp.ClientResponseError as e:
            return web.Response(status=e.status, text=e.message)
        except Exception as e:
            logger.opt(exception=e).error("Failed to proxy request")
            return web.Response(status=500)

    async def _handle_assets(self, request: web.Request) -> web.StreamResponse:
        id = request.query.get("id")
        if not id:
            return web.Response(status=400)
        identifier = Identifier.from_key(id)
        path = identifier.get_sanitized_path()
        try:
            path = safe_path_join(self._directories.assets, path)

            if not path.exists():
                return web.Response(status=404)
            return web.FileResponse(path)
        except Exception as e:
            logger.error(e)
            return web.Response(status=500)

    def run(self) -> None:
        try:
            self._loop.create_task(self.start())
            self._loop.run_forever()
        finally:
            self._loop.close()
            asyncio.run(self.shutdown())

    async def _handle_network_start(self) -> None:
        logger.info(f"Listening on {self.address}")
        try:
            await self._event.start()
        except Exception as e:
            await self.shutdown()
            self.loop.stop()
            raise e

    async def start(self) -> None:
        self._running = True
        try:
            await self._network.start()
        except Exception as e:
            await self.shutdown()
            self.loop.stop()
            raise e

    async def shutdown(self) -> None:
        self._running = False
        await self._event.stop()

    @property
    def config(self) -> Config:
        return self._config

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def address(self) -> Address:
        return self._address

    @property
    def permission_manager(self) -> PermissionManager:
        return self._security

    @property
    def directories(self) -> Directories:
        return self._directories

    @property
    def network(self) -> Network:
        return self._network

    @property
    def packet_dispatcher(self) -> ServerPacketDispatcher:
        return self._packet_dispatcher

    @property
    def endpoints(self):
        return self._endpoints

    @property
    def permissions(self):
        return self._permissions

    @property
    def tables(self):
        return self._tables

    @property
    def dashboard(self):
        return self._dashboard

    @property
    def registries(self):
        return self._registries

    @property
    def server(self):
        return self._server

    @property
    def signals(self):
        return self._signals

    @property
    def plugins(self):
        return self._plugins

    @property
    def assets(self):
        return self._assets

    @property
    def i18n(self):
        return self._i18n

    @property
    def logger(self):
        return self._logger

    @property
    def running(self) -> bool:
        return self._running

    @property
    def event(self) -> ServerEvents:
        return self._event



from omu.extension.dashboard import (
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    DASHBOARD_APP_UPDATE_PERMISSION_ID,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_SET_PERMISSION_ID,
    DASHOBARD_APP_EDIT_PERMISSION_ID,
    DASHOBARD_APP_READ_PERMISSION_ID,
)
from omu.extension.permission import PermissionType

DASHBOARD_SET_PERMISSION = PermissionType(
    DASHBOARD_SET_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "全体を管理する権限",
            "en": "Manage the dashboard",
        },
        "note": {
            "ja": "アプリが権限の管理やユーザーに確認を行うために使われます",
            "en": "Used by apps to manage permissions and confirm users",
        },
    },
)
DASHBOARD_OPEN_APP_PERMISSION = PermissionType(
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    {
        "level": "medium",
        "name": {
            "ja": "アプリを開く",
            "en": "Open an app",
        },
        "note": {
            "ja": "インストールされているアプリを起動するために使われます",
            "en": "Used to start an installed app",
        },
    },
)
DASHOBARD_APP_READ_PERMISSION = PermissionType(
    DASHOBARD_APP_READ_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "インストールされたアプリの情報を取得",
            "en": "Get Installed App Information",
        },
        "note": {
            "ja": "すでにインストールされているアプリの情報を取得するために使われます",
            "en": "Used to get information about already installed apps",
        },
    },
)
DASHOBARD_APP_EDIT_PERMISSION = PermissionType(
    DASHOBARD_APP_EDIT_PERMISSION_ID,
    {
        "level": "high",
        "name": {
            "ja": "インストールされたアプリ情報を編集",
            "en": "Edit Installed App Information",
        },
        "note": {
            "ja": "インストールされたアプリの情報を編集するために使われます",
            "en": "Used to edit information about installed apps",
        },
    },
)
DASHBOARD_APP_INSTALL_PERMISSION = PermissionType(
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    {
        "level": "high",
        "name": {
            "ja": "アプリを追加",
            "en": "Install an app",
        },
        "note": {
            "ja": "新しくアプリを追加するために使われます",
            "en": "Used to install an app",
        },
    },
)
DASHBOARD_APP_UPDATE_PERMISSION = PermissionType(
    DASHBOARD_APP_UPDATE_PERMISSION_ID,
    {
        "level": "high",
        "name": {
            "ja": "アプリ情報を更新",
            "en": "Update an app",
        },
        "note": {
            "ja": "アプリの情報を更新するために使われます",
            "en": "Used to update an app",
        },
    },
)



import os
import sys
from collections.abc import Generator
from pathlib import Path
from typing import TypedDict

import psutil


def safe_path(root_path: Path, input_path: Path) -> Path:
    """
    How to prevent directory traversal attack from Python code
    https://stackoverflow.com/a/45190125
    """
    resolved_path = root_path.joinpath(input_path).resolve()
    if not resolved_path.is_relative_to(root_path.resolve()):
        raise ValueError(f"Path {input_path} is not relative to {root_path}")
    return resolved_path.relative_to(root_path.resolve())


def safe_path_join(root: Path, *paths: Path | str) -> Path:
    return root / safe_path(root, root.joinpath(*paths))


class LaunchCommand(TypedDict):
    cwd: str
    args: list[str]


def get_launch_command() -> LaunchCommand:
    args = [sys.executable, "-m", "omuserver", *sys.argv[1:]]
    return {
        "cwd": os.getcwd(),
        "args": args,
    }


def find_processes_by_port(port: int) -> Generator[psutil.Process, None, None]:
    for connection in psutil.net_connections():
        try:
            if connection.laddr and connection.laddr.port == port:
                yield psutil.Process(connection.pid)
        except psutil.NoSuchProcess:
            pass
        except psutil.AccessDenied:
            pass


def find_processes_by_executable(
    executable: Path,
) -> Generator[psutil.Process, None, None]:
    for process in psutil.process_iter():
        try:
            if Path(process.exe()).resolve() == executable.resolve():
                yield process
        except psutil.NoSuchProcess:
            pass
        except psutil.AccessDenied:
            pass



from __future__ import annotations

import time

from omu.errors import PermissionDenied
from omu.extension.dashboard.packets import PermissionRequestPacket
from omu.extension.permission import PermissionType
from omu.extension.permission.permission_extension import (
    PERMISSION_GRANT_PACKET,
    PERMISSION_REGISTER_PACKET,
    PERMISSION_REQUEST_ENDPOINT,
    PERMISSION_REQUIRE_PACKET,
)
from omu.identifier import Identifier

from omuserver.server import Server
from omuserver.session import Session, SessionType


class PermissionExtension:
    def __init__(self, server: Server) -> None:
        server.packet_dispatcher.register(
            PERMISSION_REGISTER_PACKET,
            PERMISSION_REQUIRE_PACKET,
            PERMISSION_GRANT_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            PERMISSION_REGISTER_PACKET,
            self.handle_register,
        )
        server.packet_dispatcher.add_packet_handler(
            PERMISSION_REQUIRE_PACKET,
            self.handle_require,
        )
        server.endpoints.bind_endpoint(
            PERMISSION_REQUEST_ENDPOINT,
            self.handle_request,
        )
        self.server = server
        self.request_key = 0

    async def handle_register(
        self, session: Session, permissions: list[PermissionType]
    ) -> None:
        for permission in permissions:
            if not permission.id.is_subpath_of(session.app.id):
                msg = (
                    f"Permission identifier {permission.id} "
                    f"is not a subpart of app identifier {session.app.id}"
                )
                raise ValueError(msg)
        self.server.permission_manager.register(*permissions)

    async def handle_require(self, session: Session, permission_ids: list[Identifier]):
        if session.ready:
            raise ValueError("Session is already ready")
        if session.permission_handle.has_all(permission_ids):
            return
        if session.kind in {SessionType.PLUGIN, SessionType.DASHBOARD}:
            return

        async def task():
            permissions: list[PermissionType] = []
            for permission_id in permission_ids:
                permission = self.server.permission_manager.get_permission(
                    permission_id
                )
                if permission is None:
                    raise ValueError(f"Permission {permission_id} not registered")
                permissions.append(permission)

            request_id = self._get_next_request_key()
            accepted = await self.server.dashboard.request_permissions(
                PermissionRequestPacket(request_id, session.app, permissions)
            )
            if accepted:
                session.permission_handle.set_permissions(*[p.id for p in permissions])
                if not session.closed:
                    await session.send(PERMISSION_GRANT_PACKET, permissions)
            else:
                msg = f"Permission request denied (id={request_id})"
                raise PermissionDenied(msg)

        session.add_ready_task(task)

    async def handle_request(
        self, session: Session, permission_identifiers: list[Identifier]
    ): ...

    def _get_next_request_key(self) -> str:
        self.request_key += 1
        return f"{self.request_key}-{time.time_ns()}"



from __future__ import annotations

import abc

from loguru import logger
from omu.errors import PermissionDenied
from omu.extension.endpoint.endpoint_extension import (
    ENDPOINT_CALL_PACKET,
    ENDPOINT_ERROR_PACKET,
    ENDPOINT_RECEIVE_PACKET,
    ENDPOINT_REGISTER_PACKET,
    EndpointDataPacket,
    EndpointErrorPacket,
    EndpointType,
)
from omu.extension.endpoint.packets import EndpointRegisterPacket
from omu.helper import Coro
from omu.identifier import Identifier

from omuserver.server import Server
from omuserver.session import Session


class Endpoint(abc.ABC):
    @property
    @abc.abstractmethod
    def id(self) -> Identifier: ...

    @property
    @abc.abstractmethod
    def permission(self) -> Identifier | None: ...

    @abc.abstractmethod
    async def call(self, data: EndpointDataPacket, session: Session) -> None: ...


class SessionEndpoint(Endpoint):
    def __init__(
        self,
        session: Session,
        id: Identifier,
        permission: Identifier | None,
    ) -> None:
        self._session = session
        self._id = id
        self._permission = permission

    @property
    def id(self) -> Identifier:
        return self._id

    @property
    def permission(self) -> Identifier | None:
        return self._permission

    async def call(self, data: EndpointDataPacket, session: Session) -> None:
        if self._session.closed:
            raise RuntimeError(f"Session {self._session.app.key()} already closed")
        await self._session.send(ENDPOINT_CALL_PACKET, data)


class ServerEndpoint[Req, Res](Endpoint):
    def __init__(
        self,
        server: Server,
        endpoint: EndpointType[Req, Res],
        callback: Coro[[Session, Req], Res],
        permission: Identifier | None = None,
    ) -> None:
        self._server = server
        self._endpoint = endpoint
        self._callback = callback
        self._permission = permission

    @property
    def id(self) -> Identifier:
        return self._endpoint.id

    @property
    def permission(self) -> Identifier | None:
        return self._permission

    async def call(self, data: EndpointDataPacket, session: Session) -> None:
        if session.closed:
            raise RuntimeError("Session already closed")
        try:
            req = self._endpoint.request_serializer.deserialize(data.data)
            res = await self._callback(session, req)
            json = self._endpoint.response_serializer.serialize(res)
            await session.send(
                ENDPOINT_RECEIVE_PACKET,
                EndpointDataPacket(id=data.id, key=data.key, data=json),
            )
        except Exception as e:
            await session.send(
                ENDPOINT_ERROR_PACKET,
                EndpointErrorPacket(id=data.id, key=data.key, error=str(e)),
            )
            raise e


class EndpointCall:
    def __init__(self, session: Session, data: EndpointDataPacket) -> None:
        self._session = session
        self._data = data

    async def receive(self, data: EndpointDataPacket) -> None:
        await self._session.send(ENDPOINT_RECEIVE_PACKET, data)

    async def error(self, error: str) -> None:
        await self._session.send(
            ENDPOINT_ERROR_PACKET,
            EndpointErrorPacket(id=self._data.id, key=self._data.key, error=error),
        )


class EndpointExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        self._endpoints: dict[Identifier, Endpoint] = {}
        self._calls: dict[tuple[Identifier, int], EndpointCall] = {}
        server.packet_dispatcher.register(
            ENDPOINT_REGISTER_PACKET,
            ENDPOINT_CALL_PACKET,
            ENDPOINT_RECEIVE_PACKET,
            ENDPOINT_ERROR_PACKET,
        )
        server.packet_dispatcher.add_packet_handler(
            ENDPOINT_REGISTER_PACKET, self.handle_register
        )
        server.packet_dispatcher.add_packet_handler(
            ENDPOINT_CALL_PACKET, self.handle_call
        )
        server.packet_dispatcher.add_packet_handler(
            ENDPOINT_RECEIVE_PACKET, self.handle_receive
        )
        server.packet_dispatcher.add_packet_handler(
            ENDPOINT_ERROR_PACKET, self.handle_error
        )

    async def handle_register(
        self, session: Session, packet: EndpointRegisterPacket
    ) -> None:
        for id, permission in packet.endpoints.items():
            if not id.is_subpath_of(session.app.id):
                msg = f"App {session.app.key()} not allowed to register endpoint {id}"
                raise PermissionDenied(msg)
            self._endpoints[id] = SessionEndpoint(
                session=session,
                id=id,
                permission=permission,
            )

    def bind_endpoint[Req, Res](
        self,
        type: EndpointType[Req, Res],
        callback: Coro[[Session, Req], Res],
    ) -> None:
        if type.id in self._endpoints:
            raise ValueError(f"Endpoint {type.id.key()} already bound")
        endpoint = ServerEndpoint(
            server=self._server,
            endpoint=type,
            callback=callback,
            permission=type.permission_id,
        )
        self._endpoints[type.id] = endpoint

    def verify_permission(self, endpoint: Endpoint, session: Session):
        if endpoint.id.is_namepath_equal(session.app.id, path_length=1):
            return
        if endpoint.permission and session.permission_handle.has(endpoint.permission):
            return
        logger.warning(
            f"{session.app.key()} tried to call endpoint {endpoint.id} "
            f"without permission {endpoint.permission}"
        )
        error = (
            f"Permission denied for endpoint {endpoint.id} "
            f"with permission {endpoint.permission}"
        )
        raise PermissionDenied(error)

    async def handle_call(self, session: Session, packet: EndpointDataPacket) -> None:
        endpoint = await self._get_endpoint(packet, session)
        if endpoint is None:
            logger.warning(
                f"{session.app.key()} tried to call unknown endpoint {packet.id}"
            )
            await session.send(
                ENDPOINT_ERROR_PACKET,
                EndpointErrorPacket(
                    id=packet.id,
                    key=packet.key,
                    error=f"Endpoint {packet.id} not found",
                ),
            )
            return
        self.verify_permission(endpoint, session)

        await endpoint.call(packet, session)
        key = (packet.id, packet.key)
        self._calls[key] = EndpointCall(session, packet)

    async def handle_receive(
        self, session: Session, packet: EndpointDataPacket
    ) -> None:
        key = (packet.id, packet.key)
        call = self._calls.get(key)
        if call is None:
            await session.send(
                ENDPOINT_ERROR_PACKET,
                EndpointErrorPacket(
                    id=packet.id,
                    key=packet.key,
                    error=f"Endpoint not found {packet.id}",
                ),
            )
            return
        await call.receive(packet)

    async def handle_error(self, session: Session, packet: EndpointErrorPacket) -> None:
        key = (packet.id, packet.key)
        call = self._calls.get(key)
        if call is None:
            await session.send(
                ENDPOINT_ERROR_PACKET,
                EndpointErrorPacket(
                    id=packet.id,
                    key=packet.key,
                    error=f"Endpoint {packet.id} not found",
                ),
            )
        else:
            await call.error(packet.error)

    async def _get_endpoint(
        self, packet: EndpointDataPacket, session: Session
    ) -> Endpoint | None:
        endpoint = self._endpoints.get(packet.id)
        if endpoint is None:
            await session.send(
                ENDPOINT_ERROR_PACKET,
                EndpointErrorPacket(
                    id=packet.id,
                    key=packet.key,
                    error=f"Endpoint {packet.id} not found",
                ),
            )
            logger.warning(
                f"{session.app.key()} tried to call unconnected endpoint {packet.id}"
            )
            return
        return endpoint



from __future__ import annotations

import abc
from collections.abc import Mapping
from pathlib import Path


class TableAdapter(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def create(cls, path: Path) -> TableAdapter: ...

    @abc.abstractmethod
    async def store(self): ...

    @abc.abstractmethod
    async def load(self): ...

    @abc.abstractmethod
    async def get(self, key: str) -> bytes | None: ...

    @abc.abstractmethod
    async def get_many(self, keys: list[str] | tuple[str, ...]) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def set(self, key: str, value: bytes) -> None: ...

    @abc.abstractmethod
    async def set_all(self, items: Mapping[str, bytes]) -> None: ...

    @abc.abstractmethod
    async def remove(self, key: str) -> None: ...

    @abc.abstractmethod
    async def remove_all(self, keys: list[str] | tuple[str, ...]) -> None: ...

    @abc.abstractmethod
    async def has(self, key: str) -> bool: ...

    @abc.abstractmethod
    async def has_many(self, keys: list[str] | tuple[str, ...]) -> dict[str, bool]: ...

    @abc.abstractmethod
    async def has_all(self, keys: list[str] | tuple[str, ...]) -> bool: ...

    @abc.abstractmethod
    async def has_any(self, keys: list[str] | tuple[str, ...]) -> bool: ...

    @abc.abstractmethod
    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def fetch_range(self, start: str, end: str) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def fetch_all(self) -> dict[str, bytes]: ...

    @abc.abstractmethod
    async def first(self) -> str | None: ...

    @abc.abstractmethod
    async def last(self) -> str | None: ...

    @abc.abstractmethod
    async def clear(self) -> None: ...

    @abc.abstractmethod
    async def size(self) -> int: ...



from .asset import AssetExtension
from .endpoint import EndpointExtension
from .plugin import PluginExtension
from .registry import RegistryExtension
from .signal import SignalExtension
from .table import TableExtension

__all__ = [
    "AssetExtension",
    "EndpointExtension",
    "SignalExtension",
    "PluginExtension",
    "RegistryExtension",
    "TableExtension",
]



import io
import signal
import sys
import tracemalloc
from pathlib import Path

import click
import psutil
from loguru import logger
from omu.address import Address

from omuserver.config import Config
from omuserver.helper import find_processes_by_executable, find_processes_by_port
from omuserver.migration import migrate
from omuserver.server.omuserver import OmuServer
from omuserver.version import VERSION


def setup_logging():
    if isinstance(sys.stdout, io.TextIOWrapper):
        sys.stdout.reconfigure(encoding="utf-8")
    if isinstance(sys.stderr, io.TextIOWrapper):
        sys.stderr.reconfigure(encoding="utf-8")
    logger.add(
        "logs/{time:YYYY-MM-DD}.log",
        rotation="1 day",
        colorize=False,
        format=(
            "{time:YYYY-MM-DD HH:mm:ss} | {level: <8} | "
            "{name}:{function}:{line} - {message}"
        ),
    )


def stop_server_processes(
    port: int,
):
    executable = sys.executable
    found_processes = find_processes_by_port(port)
    if not found_processes:
        logger.info(f"No processes found using port {port}")
        return
    for process in found_processes:
        try:
            if process.exe() != executable:
                logger.warning(
                    f"Process {process.pid} ({process.name()}) is not a Python process"
                )
            logger.info(f"Killing process {process.pid} ({process.name()})")
            process.send_signal(signal.SIGTERM)
        except psutil.NoSuchProcess:
            logger.warning(f"Process {process.pid} not found")
        except psutil.AccessDenied:
            logger.warning(f"Access denied to process {process.pid}")
    for process in find_processes_by_executable(Path(executable)):
        try:
            logger.info(f"Killing process {process.pid} ({process.name()})")
            process.send_signal(signal.SIGTERM)
        except psutil.NoSuchProcess:
            logger.warning(f"Process {process.pid} not found")
        except psutil.AccessDenied:
            logger.warning(f"Access denied to process {process.pid}")


@click.command()
@click.option("--debug", is_flag=True)
@click.option("--stop", is_flag=True)
@click.option("--token", type=str, default=None)
@click.option("--token-file", type=click.Path(), default=None)
@click.option("--port", type=int, default=26423)
@click.option("--extra-trusted-origin", type=str, multiple=True)
def main(
    debug: bool,
    stop: bool,
    token: str | None,
    token_file: str | None,
    port: int,
    extra_trusted_origin: list[str],
):
    if stop:
        stop_server_processes(port)
        return

    config = Config()
    config.address = Address(
        host=None,
        port=int(port),
        secure=False,
    )

    if token:
        config.dashboard_token = token
    elif token_file:
        config.dashboard_token = Path(token_file).read_text(encoding="utf-8").strip()
    else:
        config.dashboard_token = None

    config.extra_trusted_origins = list(extra_trusted_origin)
    if config.extra_trusted_origins:
        logger.info(f"Extra trusted hosts: {config.extra_trusted_origins}")

    if debug:
        logger.warning("Debug mode enabled")
        tracemalloc.start()

    server = OmuServer(config=config)

    migrate(server)

    logger.info(f"Starting omuserver v{VERSION} on {config.address.to_url()}")
    server.run()


if __name__ == "__main__":
    setup_logging()
    try:
        main()
    except Exception as e:
        logger.opt(exception=e).error("Error running server")
        sys.exit(1)
